#import "Animator.h"

@implementation Animator

-(id)   init  :(NSMutableArray*)in_ArmatureList  :(NSMutableArray*)in_ActionList  :(NSMutableArray*)in_BVHList {
    
    if ((self = [super init])) {
        
        animationTimeMarkers     = [NSMutableArray new];

        armatureList             = in_ArmatureList;
        actionList               = in_ActionList;
        bvhList                  = in_BVHList;

        animations[typeArmature] = nil;
        animations[typeAction]   = nil;
        animations[typeBVH]      = nil;
    }

    return self;
}

-(void) uploadAnimation  :(Animation*)in_New_Animation  :(enum AnimationTypes)in_Type {
        
    NSUInteger boneCount = 0;
    lastFrameTime        = CACurrentMediaTime();
    delta                = 0.0;

    animations[in_Type] = in_New_Animation;                                             // Save new animation in its type slot; nil will disable
    
    if (!animations[in_Type]) return;                                                   // Don't bother updating a disabled or invalid animation

    switch (in_Type) {
            
        case typeArmature: {                                                            // Handle Blender-exported Armature multi-bone animation
            
            boneCount   = [armatureList count];
            
            [self calculateMsInverseBindTransforms  :armatureList[0]
                                                    :[DualQuaternion new]
                                                    :in_Type];                          // Recursively set (Model-space) inverse bind transform
            break;
        }
            
        case typeAction: {                                                              // Handle parentless/childless Action mono-bone animation
            
            boneCount   = [actionList count];                                           //TODO: Support >1 action?
            
            DualQuaternion *msBindTransform        = [actionList[0] getMsBindTransform];            // Parentless/childless Action mono-bone is in Model-space
            DualQuaternion *msInverseBindTransform = [DualQuaternion conjugate  :msBindTransform];  // Invert Model-space bindTransform to point back at origin
            
            [actionList[0] setMsInverseBindTransform  :msInverseBindTransform];         // Set (Model-space) inverse bind transform on the Action
            break;
        }
            
        case typeBVH: {                                                                 // Handle (externally imported) multi-bone BVH animation
            
            if (!(boneCount = [bvhList count])) return;                                 // Can't animate bvhList until it's been populated; bail
            
            [self calculateMsInverseBindTransforms  :bvhList[0]
                                                    :[DualQuaternion new]
                                                    :in_Type];                          // Recursively set (Model-space) inverse bind transforms
            break;
        }
            
        default:
            
            assert (false);                                                             // Unhandled AnimationType shouldn't occur; send a signal
    }
    
    for (int bone=0; bone < boneCount; bone++) animationTimeMarkers[bone] = [NSNumber numberWithFloat  :0.0];
}

-(void) calculateMsInverseBindTransforms  :(BONE*)bone  :(DualQuaternion*)in_MsParentBindTransform  :(enum AnimationTypes)in_Type {

    DualQuaternion *lsBindTransform        = [bone getLsBindTransform];                 // Get parent->bone (Local-space) bind transform (bind means un-animated)
    DualQuaternion *msBindTransform        = [DualQuaternion new];                      // For origin->bone (Model-space) bind transform
    DualQuaternion *msInverseBindTransform = [DualQuaternion new];                      // For bone->origin (Model-space) bind transform
    BONE           *childBone;
    
    msBindTransform = [DualQuaternion multiply  :in_MsParentBindTransform               // Convert Local-space bind transform to Model-space
                                                :lsBindTransform];
    
    msInverseBindTransform = [DualQuaternion conjugate  :msBindTransform];              // Invert Model-space bindTransform to point back to origin
    
    [bone setMsInverseBindTransform  :msInverseBindTransform];                          // Set the (Model-space) inverse bind transform on the bone
    
    for (NSNumber *childIndex in [bone getChildIndices]) {                              // Recursively call ourself for any Armature/BVH child bones

        switch (in_Type) {                                                              // (Note: typeAction mono-bone is handled by uploadAnimation)
                
            case typeArmature:
                
                childBone = armatureList[[childIndex intValue]];
                break;
                
            case typeBVH:
                
                childBone = bvhList[[childIndex intValue]];
                break;
                
            default:
                
                assert (false);                                                         // Unhandled AnimationType shouldn't occur; leave a trace
                break;
        }

        [self calculateMsInverseBindTransforms  :childBone  :msBindTransform  :in_Type];
    }
}

-(void) updateAnimation {
    
    if (!(animations[typeArmature] || animations[typeAction] || animations[typeBVH])) return;   // Don't bother updating if no animations are available
    
    if (![animationTimeMarkers count]) return;                                                  // Don't bother updating if no animation time markers set
    
    [self increaseAnimationTime];                                                               // Bump animationTimeMarkers, reset animations as needed
    
    [self calculateCurrentAnimationPoses];                                                      // Calculate animation poses for Armature and Actions
    
    if ((animations[typeArmature])) {                                                           // Multi-bone Armature, BVH animations need recursion
        
        [self applyPoseToChildren  :armatureList[0]  :[DualQuaternion new]  :typeArmature];     // Recursively compute poses for Armature child bones

    } else if (animations[typeBVH]) {
        
        [self applyPoseToChildren  :bvhList[0]  :[DualQuaternion new]  :typeBVH];               // Recursively compute poses for BVH child bones
    }
    
    [self updateAnimationTiming];                                                               // Update animation timing variables (delta, lastFrameTime)
}

-(void) increaseAnimationTime {
    
    double timeLimit;
    
    NSUInteger boneCount;

    if (animations[typeArmature] || animations[typeBVH]) {                              // Decide whether to use Armature or BVH anims
        
        boneCount = (animations[typeArmature]) ? [armatureList count]                   // If available, use Armature animation count
                                               : [bvhList count];                       // Otherwise, use any BVH animation count
        
    } else boneCount = [actionList count];                                              // Neither Armature nor BVH available, use Action
    
    for (int bone=0; bone < boneCount; bone++) {                                        // For all bones in Armature/BVH, or Action's single bone
        
        animationTimeMarkers[bone] = [NSNumber numberWithDouble  :[animationTimeMarkers[bone] doubleValue] + delta];  // Increase bone time marker
        
        if (animations[typeArmature] || animations[typeBVH]) {                          // Decide whether to use Armature or BVH anims
            
            timeLimit = (animations[typeArmature]) ? [[animations[typeArmature] getBoneAnimationLengths][bone] doubleValue]  // Use Armature animation lengths
                                                   : [[animations[typeBVH]      getBoneAnimationLengths][bone] doubleValue]; // Otherwise, use BVH lengths
            
        } else timeLimit = [[animations[typeAction] getBoneAnimationLengths][bone] doubleValue];    // Neither Armature nor BVH available, use Action

        if ([animationTimeMarkers[bone] doubleValue] >= timeLimit) {                    // Reset if bone's time marker reaches time limit
            
            double remainder = fmod ([animationTimeMarkers[bone] doubleValue], timeLimit);
            
            animationTimeMarkers[bone] = [NSNumber numberWithDouble  :remainder];
        }
    }
}

-(void) calculateCurrentAnimationPoses {
    
    NSMutableArray  *boneKeyFrames      = nil;                                              // Array of bones BoneKeyFrame objects
    NSMutableArray  *neighbouringFrames = nil;                                              // Array of two neighbouring KeyFrames
    NSMutableArray  *boneAnimations     = nil;                                              // Array of all current bone animations
    
    if  (animations[typeArmature]) boneAnimations = [animations[typeArmature] getBoneAnimations];
    if  (animations[typeAction])   boneAnimations = [animations[typeAction]   getBoneAnimations];
    if  (animations[typeBVH])      boneAnimations = [animations[typeBVH]      getBoneAnimations];

    for (int bone=0; bone < [boneAnimations count]; bone++) {
        
        NSNumber *currentTimeMarker = animationTimeMarkers[bone];                           // Get this bones current animation time marker

        boneKeyFrames = boneAnimations[bone];                                               // Get array of this bones BoneKeyFrame objects
        
        if ([boneKeyFrames count] == 0) continue;                                           // Bypass bones that aren't actually animated
       
        neighbouringFrames = [self getPreviousAndNextFrames  :boneKeyFrames                 // Find neighbouring KeyFrames by timeStamps
                                                             :currentTimeMarker];
        
        float progression = [self calculateProgression  :neighbouringFrames[0]              // Compute current interpolation progression
                                                        :neighbouringFrames[1]
                                                        :currentTimeMarker];
        
        DualQuaternion *transform = [DualQuaternion interpolate  :[neighbouringFrames[0] getBoneTransform]  // Previous KeyFrame transform
                                                                 :[neighbouringFrames[1] getBoneTransform]  // and next KeyFrame transform
                                                                 :progression];                             // Interpolate at progression
        
        if  (animations[typeArmature]) [armatureList[bone] setLsAnimatedPoseTransform  :transform]; // Armature/BVH poses are Local-space
        if  (animations[typeAction])   [actionList[bone]   setMsAnimatedPoseTransform  :transform]; // While Action poses are Model-space
        if  (animations[typeBVH])      [bvhList[bone]      setLsAnimatedPoseTransform  :transform]; // Armature/BVH poses are Local-space
    }
}

-(void) applyPoseToChildren  :(BONE*)bone  :(DualQuaternion*)in_MsParentTransform  :(enum AnimationTypes)in_Type {
    
    if (!bone) return;                                                                      // No point trying to animate an invalid bone/joint
    
    DualQuaternion *lsAnimatedPoseTransform = [bone getLsAnimatedPoseTransform];            // Get current Local-space animated pose transform

    DualQuaternion *msCurrentTransform = [DualQuaternion multiply  :in_MsParentTransform    // Calculate Model-space animated pose transform
                                                                   :lsAnimatedPoseTransform];
    BONE *childBone;
    
    for (NSNumber *childIndex in [bone getChildIndices]) {                                  // Recursively call ourself for any child bones:
        
        if (in_Type == typeArmature) childBone = armatureList[[childIndex intValue]];
        
        else if (in_Type == typeBVH) childBone = bvhList[[childIndex intValue]];
        
        else assert (false);                                                                // Trap on an unexpected type (e.g. typeAction)
        
        [self applyPoseToChildren  :childBone  :msCurrentTransform  :in_Type];
    }
    
    msCurrentTransform = [DualQuaternion multiply  :msCurrentTransform                      // Calculate Model-space 'Bind-to-Pose' transform
                                                   :[bone getMsInverseBindTransform]];      // (https://www.youtube.com/watch?v=cieheqt7eqc)
    
    [bone setMsAnimatedPoseTransform  :msCurrentTransform];                                 // Set bone's Model-space animated pose Dual Quaternion
}

-(void) updateAnimationTiming {

    double currentFrameTime = CACurrentMediaTime();
    
    delta         = (currentFrameTime - lastFrameTime);
    lastFrameTime = currentFrameTime;
}

-(NSMutableArray*) getPreviousAndNextFrames  :(NSMutableArray*)in_KeyFrames  :(NSNumber*)in_CurrentTimeMarker {
    
    float timeMarker = [in_CurrentTimeMarker floatValue];
    
    BoneKeyFrame *previousFrame = in_KeyFrames[0];
    BoneKeyFrame *nextFrame     = in_KeyFrames[0];

    for (int i = 1; i < [in_KeyFrames count]; i++) {                // (This simple O(n) for-loop takes 60% less time than indexOfObjectPassingTest in a block)
                                                                    // (Can't use a Binary Search to find nextFrame, since in_Keyframes isn't a simple NSArray)
        nextFrame = in_KeyFrames[i];

        if ([nextFrame getTimeStamp] >= timeMarker) break;

        previousFrame = in_KeyFrames[i];
    }

    NSMutableArray *outputKeyFrames = [[NSMutableArray alloc] initWithCapacity  :2];

    [outputKeyFrames insertObject  :previousFrame atIndex:0];
    [outputKeyFrames insertObject  :nextFrame     atIndex:1];

    return outputKeyFrames;
}

-(float) calculateProgression  :(BoneKeyFrame*)in_PreviousFrame  :(BoneKeyFrame*)in_NextFrame  :(NSNumber*)in_CurrentTimeMarker {

    float   currentTime, totalTime;

    currentTime  = [in_CurrentTimeMarker floatValue] - [in_PreviousFrame getTimeStamp];

    totalTime    = [in_NextFrame getTimeStamp] - [in_PreviousFrame getTimeStamp];

    return  ((totalTime > 0.0) ? (currentTime / totalTime)  : 0.0);
}

-(Animation*) getAnimation  :(enum AnimationTypes)in_Type { return animations[in_Type]; }

@end
