#import "DualQuaternion.h"

using namespace std;

@implementation DualQuaternion

-(id)   init {
    
    if ((self = [super init])) {                                        // Dual Quaternion Identity, per Kenwright(1), Equation 24
        
        m_real = [[Quaternion alloc] init  :0.0  :0.0  :0.0  W:1.0];    // Rotation part as (0,0,0, W=1)
        m_dual = [[Quaternion alloc] init  :0.0  :0.0  :0.0  W:0.0];    // Translation part (0,0,0, W=0)
    }
    
    return self;
}

-(id)   init  :(Quaternion*)in_Real  :(Quaternion*)in_Dual {
    
    if ((self = [super init])) {
        
        assert ((in_Real != nil) && (in_Dual != nil));

        m_real = [Quaternion normalize  :in_Real];
        m_dual = in_Dual;
    }
    
    return self;
}

-(id)   initFromGLM  :(glm::mat4)in_glmMatrix {

    if ((self = [super init])) {                                                                // Use Eigen since bug in glm::quat_cast from commit c6606998
        
        using EigenMatrix = Eigen::Map<Eigen::Matrix4f>;                                        // Bring Eigen::Matrix4f into scope (saves a little typing)
       
        EigenMatrix eMatrix4 = EigenMatrix (glm::value_ptr (in_glmMatrix));                     // Convert glm::mat4 to Eigen::Matrix4f

        return [self initFromEigenMatrix  :eMatrix4];
    }

    return self;
}

-(id)   initFromMatrix  :(Matrix4f*)in_Matrix {

    if ((self = [super init])) {                                                                // Use Eigen since bug in glm::quat_cast from commit c6606998
        
        using EigenMatrix = Eigen::Map<Eigen::Matrix4f>;                                        // Bring Eigen::Matrix4f into scope (saves a little typing)
       
        EigenMatrix eMatrix4 = EigenMatrix ([in_Matrix getPtr]);                                // Convert MeshP2::Matrix4f to Eigen::Matrix4f
        
        return [self initFromEigenMatrix  :eMatrix4];
    }

    return self;
}

-(id)   initFromEigenMatrix  :(Eigen::Map<Eigen::Matrix4f>)in_eMatrix4 {

    if ((self = [super init])) {                                                                // Use Eigen since bug in glm::quat_cast from commit c6606998

        Eigen::Quaternionf eRotation = Eigen::Quaternionf (in_eMatrix4.topLeftCorner<3, 3>());  // Extract the 3x3 rotational sub-matrix

        Eigen::Quaternionf eTranslation = Eigen::Quaternionf (in_eMatrix4.col(3));              // Extract the translation vector
        
        Eigen::Quaternionf eReal = eRotation;
                           eReal.normalize();                                                   // Create a pure (i.e. orthonormalized) rotational quaternion
        
        Eigen::Quaternionf eDual = Eigen::Quaternionf (0.0, eTranslation.x(), eTranslation.y(), eTranslation.z());  // Note that ordering is: W, then X, Y, Z
                           eDual = 0.5 * (eDual * eReal).coeffs();                              // Create a translation quaternion (Kenwright(1) Equation 25)

        m_real = [[Quaternion alloc] init  :eReal.x()  :eReal.y()  :eReal.z()  W:eReal.w()];
        m_dual = [[Quaternion alloc] init  :eDual.x()  :eDual.y()  :eDual.z()  W:eDual.w()];
    }

    return self;
}

-(id)   initFromAngleAxis  :(float)in_AngleDegrees  :(Vector3f*)in_Axes {

    if ((self = [super init])) {
        
        m_real  = [[Quaternion alloc] initFromAngleAxis  :in_AngleDegrees  :[in_Axes normalize]];   // Ensure rotation axes are normalized
        
        m_dual  = [[Quaternion alloc] init  :0.0  :0.0  :0.0  W:0.0];                               // Kenwright(1): a rotation-only DQ has no translation part
    }
    
    return self;
}

-(NSString*) toString { return [NSString stringWithFormat  :@"real: %@\tdual: %@", [m_real toString], [m_dual toString]]; }

+(Quaternion*)  getRotation  :(DualQuaternion*)dq {
    
    assert (dq != nil);

    return dq->m_real;
}

+(Vector3f*)    getTranslation  :(DualQuaternion*)dq {
    
    assert (dq != nil);

    glm::dualquat glmDQ = [self toGLMdualquat  :dq];

    glm::quat glmXlate  = glm::quat ((2.0f * glmDQ.dual) * glm::conjugate (glmDQ.real));        // Per Kenwright(1) Equation 29

    return [[Vector3f alloc] init  :glmXlate.x  :glmXlate.y  :glmXlate.z];
}

+(Quaternion*)  getReal  :(DualQuaternion*)dq {
    
    assert (dq != nil);

    return dq->m_real;
}

+(Quaternion*)  getDual  :(DualQuaternion*)dq {
    
    assert (dq != nil);

    return dq->m_dual;
}

+(DualQuaternion*)  normalize  :(DualQuaternion*)dq {
    
    assert (dq != nil);
    
    return [[DualQuaternion alloc] init  :[Quaternion   normalize  :[DualQuaternion getReal  :dq]]
                                         :[DualQuaternion getDual  :dq]];       //Kenwright(1) Appendix 'normalize' includes Dual, but that seems wrong
}

+(DualQuaternion*)  conjugate  :(DualQuaternion*)dq {
    
    assert (dq != nil);

    DualQuaternion *dest    = [DualQuaternion new];
    
    glm::dualquat glmDQ     = [self toGLMdualquat  :dq];

    glm::dualquat glmResult = glm::inverse (glmDQ);

    dest->m_real = [[Quaternion alloc] init  :glmResult.real.x  :glmResult.real.y  :glmResult.real.z  W:glmResult.real.w];
    dest->m_dual = [[Quaternion alloc] init  :glmResult.dual.x  :glmResult.dual.y  :glmResult.dual.z  W:glmResult.dual.w];
    
    return dest;
}

+(DualQuaternion*)  multiply  :(DualQuaternion*)lhs  :(DualQuaternion*)rhs {

    assert ((lhs != nil) && (rhs != nil));

// We can't use Kenwright(1) Appendix: DualQuaternion_c operator *(DualQuaternion_c lhs, DualQuaternion_c rhs) as gimpiness results

    glm::dualquat glmResult = ([DualQuaternion toGLMdualquat  :lhs]) * ([DualQuaternion toGLMdualquat  :rhs]);  // Order's important
    
    Quaternion *resultReal  = [[Quaternion alloc] init  :glmResult.real.x  :glmResult.real.y  :glmResult.real.z  W:glmResult.real.w];
    Quaternion *resultDual  = [[Quaternion alloc] init  :glmResult.dual.x  :glmResult.dual.y  :glmResult.dual.z  W:glmResult.dual.w];
    
    return [[DualQuaternion alloc] init  :resultReal  :resultDual];
}

+(DualQuaternion*)  interpolate  :(DualQuaternion*)fromDQ  :(DualQuaternion*)toDQ  :(float)progression {

    assert ((fromDQ != nil) && (toDQ != nil));
    assert ((progression >= 0.0) && (progression <= 1.0));

//  Note: For some reason, we can't use individual [Quaternion interpolate] calls for Real and Dual quaternions as gimpiness results

    glm::dualquat glmLERP = glm::lerp ([self toGLMdualquat  :fromDQ], [self toGLMdualquat  :toDQ], progression);

    return [[DualQuaternion alloc] init  :[[Quaternion alloc] init  :glmLERP.real.x  :glmLERP.real.y  :glmLERP.real.z  W:glmLERP.real.w]
                                         :[[Quaternion alloc] init  :glmLERP.dual.x  :glmLERP.dual.y  :glmLERP.dual.z  W:glmLERP.dual.w]];
}

+(glm::dualquat)  toGLMdualquat  :(DualQuaternion*)dq {
    
    assert (dq != nil);

    Quaternion *dqReal  = [DualQuaternion getReal  :dq];
    Quaternion *dqDual  = [DualQuaternion getDual  :dq];
    
    return glm::dualquat (glm::quat (dqReal.getW,  dqReal.getX,  dqReal.getY,  dqReal.getZ),    // glm::quat ordering is W 1st then X, Y, Z
                          glm::quat (dqDual.getW,  dqDual.getX,  dqDual.getY,  dqDual.getZ));   // which is the exact opposite of glm::vec4
}

@end

//  (1) Benjamin Kenwright: Dual Quaternions - From Classical Mechanics to Computer Graphics and Beyond
//      (https://xbdev.net/misc_demos/demos/dual_quaternions_beyond/paper.pdf)
