#import "Math_Routines.h"

@implementation MeshP2 (Math_Routines)

-(GLfloat)      Angle  :(SURFACE*)surface  :(GLuint)vertex {    // Returns the angle of a triangle at the specified vertex
    
	double result;
	double a, b, c;
    
	a = b = c = 0;
    
	if  (vertex == [surface getP0]) {           // Calculate length of sides a, b and c when ∠A is P0
        
        a = [self Distance  :VertexList[[surface getP1]]  :VertexList[[surface getP2]]];
        b = [self Distance  :VertexList[[surface getP0]]  :VertexList[[surface getP1]]];
        c = [self Distance  :VertexList[[surface getP0]]  :VertexList[[surface getP2]]];
        
    } else if (vertex == [surface getP1]) {     // ... or, when ∠A is P1
        
        a = [self Distance  :VertexList[[surface getP2]]  :VertexList[[surface getP0]]];
        b = [self Distance  :VertexList[[surface getP2]]  :VertexList[[surface getP1]]];
        c = [self Distance  :VertexList[[surface getP0]]  :VertexList[[surface getP1]]];
        
    } else if (vertex == [surface getP2]) {     // ... or, when ∠A is P2
        
        a = [self Distance  :VertexList[[surface getP1]]  :VertexList[[surface getP0]]];
        b = [self Distance  :VertexList[[surface getP2]]  :VertexList[[surface getP0]]];
        c = [self Distance  :VertexList[[surface getP1]]  :VertexList[[surface getP2]]];
    }
    
	result  = (b*b + c*c - a*a);			    // Compute the angle at this vertex as:  cosA = (b^2 + c^2 - a^2) / 2bc

	result /= (2*b*c);

	result	= acos (result);				    // Take the inverse of cosA to get the angle in radians
	
	result *= (180.0/M_PI);
	
	return (GLfloat) result;				    // Triangle ∠A at the requested vertex in degrees
}

-(Vector3f*)    Tangent  :(SURFACE*)triangle  :(GLuint)vertex { // Returns a Tangent Vector for a given triangle and vertex
                                                                // Based on http://ogldev.atspace.co.uk/www/tutorial26/tutorial26.html
	VERTEX		*v0, *v1, *v2;
	
    Vector3f	*tangent = nil;

	v0 = v1 = v2 = nil;
	
	if (vertex == [triangle getP0]) {
		
		v0 = VertexList[[triangle getP0]];
		v1 = VertexList[[triangle getP1]];
		v2 = VertexList[[triangle getP2]];
		
	} else if (vertex == [triangle getP1]) {
		
		v0 = VertexList[[triangle getP1]];
		v1 = VertexList[[triangle getP2]];
		v2 = VertexList[[triangle getP0]];
		
	} else if (vertex == [triangle getP2]) {
		
		v0 = VertexList[[triangle getP2]];
		v1 = VertexList[[triangle getP0]];
		v2 = VertexList[[triangle getP1]];
	}
	
    Vector3f *Edge1 = [[Vector3f alloc] init  :[v1 getX]-[v0 getX]    :[v1 getY]-[v0 getY]    :[v1 getZ]-[v0 getZ]];
    Vector3f *Edge2	= [[Vector3f alloc] init  :[v2 getX]-[v0 getX]    :[v2 getY]-[v0 getY]    :[v2 getZ]-[v0 getZ]];
	
	float DeltaU1	= [v1 getU]-[v0 getU];
	float DeltaV1	= [v1 getV]-[v0 getV];
	
	float DeltaU2	= [v2 getU]-[v0 getU];
	float DeltaV2	= [v2 getV]-[v0 getV];
	
	float r			= DeltaU1 * DeltaV2 - DeltaU2 * DeltaV1;
	
	if	(!r) return tangent;							        // If there are no texcoords for this vertex, just return a null tangent vector
	
	float f			= 1.0f / r;
	
    float x	= f * (DeltaV2 * [Edge1 getX] - DeltaV1 * [Edge2 getX]);
	float y	= f * (DeltaV2 * [Edge1 getY] - DeltaV1 * [Edge2 getY]);
	float z	= f * (DeltaV2 * [Edge1 getZ] - DeltaV1 * [Edge2 getZ]);

    tangent = [[Vector3f alloc] init  :x  :y  :z];
	
    return tangent;
}

-(GLfloat)      Cos_angle  :(SURFACE*)s1  :(SURFACE*)s2 {       // Returns the Dot Product of two vectors; in this case two surface normals
/*
    Since both surface normals are already normalized as unit vectors, their dot product is the same as the cosine of the angle between the normals.
 
    See "Scalar projection" at https://en.wikipedia.org/wiki/Dot_product
 */
    return [s1 getNX] * [s2 getNX] + [s1 getNY] * [s2 getNY] + [s1 getNZ] * [s2 getNZ];
}

-(double)       Distance  :(VERTEX*)v1  :(VERTEX*)v2 {          // Returns the distance between two 3D points
    
    double dx, dy, dz;
	
    dx = fabs (([v2 getX] - [v1 getX]) * ([v2 getX] - [v1 getX]));
    dy = fabs (([v2 getY] - [v1 getY]) * ([v2 getY] - [v1 getY]));
    dz = fabs (([v2 getZ] - [v1 getZ]) * ([v2 getZ] - [v1 getZ]));
    
    return sqrt (dx + dy + dz);
}

@end
