#import <OpenGL/gl3.h>

#define GLM_ENABLE_EXPERIMENTAL
#import <glm/glm.hpp>
#import <glm/ext.hpp>
#import <glm/gtx/dual_quaternion.hpp>

#import "Constants.h"
#import "VERTEX.h"                              // Importing VERTEX.h allows MeshP2 calling apps to process vertices (File Inspector: set 'Target Membership' to 'public')
#import "BONE.h"                                // Importing BONE.h allows callers to process Armature-based bones, e.g. setting per-keyframe animation transforms
#import "ACTION.h"                              // Importing ACTION.h allows callers to process Action-based animations i.e. non-Armature based animations
#import "Animator.h"                            // Importing Animator allows callers to set/start/stop animation keyframes for individual bones
#import "Matrix4fDecomposed.h"                  // Importing Matrix4fDecomposed.h allows callers to decompose a glm::mat4 into components

@interface MeshP2 : NSObject {
    
    NSMutableArray *VertexList;			        // Array of VERTEXs		 - indexed per vertex;    positions, normals, texcoords, tangents and vertex flags
    NSMutableArray *SurfaceList;			    // Array of SURFACEs     - indexed per vertex;    surfaces at, or sharing, an individual vertex
    NSMutableArray *PolygonList;                // Array of POLYGONLISTs - indexed per material;  material properties and linked surfaces
    NSMutableArray *ArmatureList;               // Array of BONEs        - indexed per bone;      Blender's Armature defined animations
    NSMutableArray *ActionList;                 // Array of ACTIONs      - indexed per action;    Blender's Action defined animations
    NSMutableArray *BVHList;                    // Array of BONEs        - indexed per 'joint';   BVH file imported joint animations

    Animator       *animator;                   // Manages any animation keyframes added by MeshP2 calling apps, e.g. DisplayP2

    id				Scene;
    FILE		   *InFile;
    NSString       *MeshFileName;
    
    GLuint          VertexShader;               // Vertex shader ID
    GLuint          FragmentShader;             // Fragment shader ID
    GLuint          ShaderProgram;              // Compiled shader program ID
    GLuint          meshVAO;                    // Unique ID of the Vertex Array Object - an array of VBOs used to cache model parameters
    GLuint          meshVBO;                    // Unique ID of the Vertex Buffer Object (VBO) for all mesh vertices, normals, texcoords amd tangents
    GLuint          vboSize;
    GLfloat        *vboBuffer;                  // Client-side buffer storing all vertex positions, normals, texcoords and tangents before VBO upload
    
    GLfloat         cosineThreshold;            // Cosine of threshold angle which determines when to split vertices shared between multiple surfaces
    GLfloat         max_X, max_Y, max_Z;        // Maximum X, Y and Z values found in the mesh object; useful for callers to initially place a camera

    Matrix4f       *ModelViewMatrix;            // Current ModelView matrix, updated by caller when a mesh is rotated, scaled, zoomed, etc. and sent to vertex shader
    Matrix4f       *ProjectionMatrix;           // Current Projection matrix, updated by caller when a mesh is loaded or a drawing window is reshaped; sent to shader
    
    Vector3f       *ScalingVector;              // Current Scaling vector, imported from .P2 mesh as CHUNK_MASTER_SCALE; currently unused
    
    bool            force2Sided;                // Flag forces 2-Sided materials
    bool            P2_Logging_Enabled;         // Flag enables debug message logging
    bool            tooManyBonesWarning;        // Flag set if 'Too Many Bones' warning issued

    Vector4f       *globalLightPosition;        // Global light properties for the Okudagram... //TODO: Add more lights?
    Vector4f       *globalLightAmbient;
    Vector4f       *globalLightDiffuse;
    Vector4f       *globalLightSpecular;
    
    GLfloat         globalLightConstantAttenuation;
    GLfloat         globalLightLinearAttenuation;
    GLfloat         globalLightQuadraticAttenuation;
    
    GLint           locModelViewMatrix;         // Shader Uniform Locations...
    GLint           locProjectionMatrix;
    
    GLint           locGlobalLightPosition;
    GLint           locGlobalLightAmbient;
    GLint           locGlobalLightDiffuse;
    GLint           locGlobalLightSpecular;
    
    GLint           locGlobalLightConstantAttenuation;
    GLint           locGlobalLightLinearAttenuation;
    GLint           locGlobalLightQuadraticAttenuation;
    
    GLint           locMaterialLightAmbient;
    GLint           locMaterialLightDiffuse;
    GLint           locMaterialLightSpecular;
    
    GLint           locMaterialShininess;
    GLint           locMaterialAlpha;
    
    GLboolean       locTextureMap1Flag;
    GLint           locTextureMap1Unit;
    GLboolean       locTextureMap2Flag;
    GLint           locTextureMap2Unit;
    
    GLboolean       locBumpMapFlag;
    GLint           locBumpMapUnit;

    GLint           locAnimatedBoneCount;
    GLint           locBoneTransforms[MAX_BONES];
}

-(id)   init                 :(GLfloat)in_threshold  :(bool)in_force2sided  :(bool)in_logging;

-(bool) ParseP2              :(NSString*)in_Filename;                                       // Parses to P2* data structures, converting to OpenGL
-(void) DisplayMesh;                                                                        // Displays mesh onto current OpenGL Rendering Context
-(void) UploadBVHList        :(NSMutableArray*)in_newBVHList;                               // Accepts new BVHList parsed/sent from MeshP2 callers
-(void) UploadAnimation      :(Animation*)in_NewAnimation  :(enum AnimationTypes)in_Type;   // Accepts new Animation keyframes from MeshP2 callers

-(void) destroyVAOVBOandIBO;
-(void) dealloc;

-(NSMutableArray*) getVertexList;
-(NSMutableArray*) getArmatureList;
-(NSMutableArray*) getActionList;
-(NSMutableArray*) getBVHList;
-(Matrix4f*)       getModelViewMatrix;
-(NSString*)       getMeshFilename;
-(GLfloat)         getMax_X;
-(GLfloat)         getMax_Y;
-(GLfloat)         getMax_Z;

-(void) SetModelViewMatrix      :(Matrix4f*)in_ModelView_Matrix;
-(void) SetProjectionMatrix     :(Matrix4f*)in_Projection_Matrix;

-(void) SetLightProperties      :(Vector4f*)in_Position
                                :(Vector4f*)in_Ambient  :(Vector4f*)in_Diffuse  :(Vector4f*)in_Specular
                                :(GLfloat)in_ConstantAttenuation  :(GLfloat)in_LinearAttenuation  :(GLfloat)in_QuadraticAttenuation;

+(void) P2Log                   :(bool)in_Enabled  :(NSString*)format, ...;

@end
