#import "Constants.h"
#import "MeshP2.h"
#import "File_Routines.h"
#import "OpenGL_Routines.h"

@implementation MeshP2

-(id)   init  :(GLfloat)in_threshold  :(bool)in_force2sided  :(bool)in_logging {
	
    if	((self = [super init]))	{
        
        VertexList           = nil;
        PolygonList          = nil;
        SurfaceList          = nil;
        ArmatureList         = nil;
        ActionList           = nil;
        BVHList              = nil;
        Scene                = nil;
		InFile				 = nil;
        MeshFileName         = nil;
        animator             = nil;
        vboBuffer            = nil;
        vboSize              = 0;
        meshVAO              = 0;
        meshVBO              = 0;
        VertexShader         = 0;
        FragmentShader       = 0;
        ShaderProgram        = 0;
        force2Sided          = in_force2sided;
        cosineThreshold      = in_threshold;
        P2_Logging_Enabled   = in_logging;
        tooManyBonesWarning  = false;
        ModelViewMatrix      = [Matrix4f new];
        ProjectionMatrix     = [Matrix4f new];
    }
    
	return self;
}

-(bool) ParseP2  :(NSString*)in_Filename {
    
    NSException *exReadScene = [NSException exceptionWithName  :@"ParseP2 Error"  reason:@"Failed to Read Scene"  userInfo:nil];
    NSException *exOpenFile  = [NSException exceptionWithName  :@"ParseP2 Error"  reason:@"Failed to Open File"   userInfo:nil];
    NSException *exSeekError = [NSException exceptionWithName  :@"ParseP2 Error"  reason:@"Seek Error on File"    userInfo:nil];
    NSException *exSizeError = [NSException exceptionWithName  :@"ParseP2 Error"  reason:@"Size Error on File"    userInfo:nil];

    @synchronized (self) {                                                                          // Delay access by any concurrent threads, e.g. splash screens

        FILE  *inf;
        long  size;

        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ParseP2: filename: %@", in_Filename];

        @try {

            if (!(inf = fopen ([in_Filename UTF8String], "rb"))) {
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ParseP2: fopen error %d (%s) for filename: %@", errno, strerror (errno), in_Filename];

                @throw exOpenFile;
            }
            
            if (fseek (inf, 0, SEEK_END))    { fclose (inf); @throw exSeekError; }
            
            if ((size = ftell (inf)) == -1L) { fclose (inf); @throw exSizeError; }
            
            rewind (inf);

            MeshFileName    = in_Filename;                                                          // Save the mesh file path for possible use by LoadTextureFile

            Scene           = [P2Scene new];
            
            if (![self ReadScene  :inf  :(int) size]) {                                             // Parse the .P2 mesh into local P2* data structures. If error, clean and return
                
                fclose (inf);
                
                Scene       = nil;
                VertexList  = nil;
                PolygonList = nil;
                SurfaceList = nil;
                ArmatureList= nil;
                ActionList  = nil;
                BVHList     = nil;
                animator    = nil;

                @throw exReadScene;
            }
            
            fclose (inf);
            
            [self Convert_to_OpenGL];                                                               // Migrate internal P2* data to publicly accessible equivalents (e.g. VertexList)
            
            animator = [[Animator alloc] init  :ArmatureList  :ActionList  :BVHList];               // Wrap {Armature|Action|BVH}Lists in animator object, before receiving animations
        
        } @catch (NSException *ex) {
            
            NSAlert *alert = [NSAlert new];

            [alert setAlertStyle       :NSAlertStyleWarning];
            [alert setMessageText      :@"DisplayP2 Error!"];
            [alert setInformativeText  :@"Failed @ ParseP2"];
            [alert addButtonWithTitle  :@"OK"];
           
            NSArray *backtrace = [ex callStackSymbols];
            
            [MeshP2 P2Log  :P2_Logging_Enabled  :@"Exception raised in MeshP2->ParseP2 e = %@\n\n Backtrace:\n%@", ex, backtrace];

            [alert runModal];

            return false;
        }

        return true;
    }
}

-(NSMutableArray*) getVertexList      { return VertexList; }
-(NSMutableArray*) getArmatureList    { return ArmatureList; }
-(NSMutableArray*) getActionList      { return ActionList; }
-(NSMutableArray*) getBVHList         { return BVHList; }
-(Matrix4f*)       getModelViewMatrix { return ModelViewMatrix; }
-(NSString*)       getMeshFilename    { return MeshFileName; }
-(GLfloat)         getMax_X           { return max_X; }
-(GLfloat)         getMax_Y           { return max_Y; }
-(GLfloat)         getMax_Z           { return max_Z; }

-(void) SetModelViewMatrix      :(Matrix4f*)in_ModelView_Matrix  { ModelViewMatrix  = in_ModelView_Matrix; }
-(void) SetProjectionMatrix     :(Matrix4f*)in_Projection_Matrix { ProjectionMatrix = in_Projection_Matrix; }

-(void) SetLightProperties   :(Vector4f*)in_Position
                             :(Vector4f*)in_Ambient  :(Vector4f*)in_Diffuse  :(Vector4f*)in_Specular
                             :(GLfloat)in_ConstantAttenuation  :(GLfloat)in_LinearAttenuation  :(GLfloat)in_QuadraticAttenuation {
/*
    Set the Global lighting properties, as compared to individual Material lighting properties set during Display //TODO: Add more lights
*/
    globalLightPosition             = in_Position;
    globalLightAmbient              = in_Ambient;
    globalLightDiffuse              = in_Diffuse;
    globalLightSpecular             = in_Specular;

    globalLightConstantAttenuation  = in_ConstantAttenuation;
    globalLightLinearAttenuation    = in_LinearAttenuation;
    globalLightQuadraticAttenuation = in_QuadraticAttenuation;
}

+(void) P2Log  :(bool)in_Enabled  :(NSString *)format, ... {
    
    if (in_Enabled) {
        
        va_list     args;
        va_start    (args, format);
        NSLogv      (format, args);
//      printf      ("%s\n", [[NSString stringWithFormat  :format, args] UTF8String]);              // printf allows log messages > 1024 bytes
        va_end      (args);
    }
}

-(void) DisplayMesh {
        
    if (glIsProgram  (ShaderProgram)) {                                                             // If valid Vertex/Fragment shaders compiled and linked
        
        glUseProgram (ShaderProgram);                                                               // Use Vertex/Fragment shaders in the rendering state

        glBindVertexArray (meshVAO);                                                                // Bind the VAO and its internally associated VBO

        if (([animator getAnimation  :typeArmature]) || ([animator getAnimation  :typeAction]) || ([animator getAnimation  :typeBVH])) {
            
            [animator updateAnimation];                                                             // Calculate animated pose transforms

            [self loadBoneTransformUniforms];                                                       // Load BoneTransform array uniforms

        } else glUniform1i  (locAnimatedBoneCount, 0);                                              // Otherwise, tell shader to relax

        glUniformMatrix4fv  (locModelViewMatrix,  1, GL_FALSE, &(ModelViewMatrix->m00));            // Set the MVP matrices as shader uniforms
        glUniformMatrix4fv  (locProjectionMatrix, 1, GL_FALSE, &(ProjectionMatrix->m00));

        glUniform4fv        (locGlobalLightPosition, 1, globalLightPosition.getXptr);               // Load Global lighting vectors into uniforms
        glUniform4fv        (locGlobalLightAmbient,  1, globalLightAmbient.getXptr);
        glUniform4fv        (locGlobalLightDiffuse,  1, globalLightDiffuse.getXptr);
        glUniform4fv        (locGlobalLightSpecular, 1, globalLightSpecular.getXptr);
        
        glUniform1f         (locGlobalLightConstantAttenuation,  globalLightConstantAttenuation);   // Load Global lighting attenuation scalars into uniforms
        glUniform1f         (locGlobalLightLinearAttenuation,    globalLightLinearAttenuation);
        glUniform1f         (locGlobalLightQuadraticAttenuation, globalLightQuadraticAttenuation);

        for	(POLYGONLIST *poly_entry in PolygonList) {                                              // For every material in PolygonList:
            
            if (glIsBuffer ([poly_entry getMaterialIBO])) {                                         // If there's a valid Index Buffer Object (IBO):
                
                glBindBuffer        (GL_ELEMENT_ARRAY_BUFFER, [poly_entry getMaterialIBO]);         // Bind the IBO, indexing into the VBO for the material's vertices...
                                                                                                    // ... normals, tangents, texcoords, bone indices & weights
                glUniform4fv        (locMaterialLightAmbient,  1, [poly_entry getAmbient]);         // Load Material lighting vectors into uniforms...
                glUniform4fv        (locMaterialLightDiffuse,  1, [poly_entry getDiffuse]);
                glUniform4fv        (locMaterialLightSpecular, 1, [poly_entry getSpecular]);
                
                glUniform1f         (locMaterialShininess, [poly_entry getShininess]);              // Material lighting scalars...
                glUniform1f         (locMaterialAlpha,     [poly_entry getAlpha]);
                
                glActiveTexture	    (GL_TEXTURE0);												    // Activate texture unit 0 for 1st TextureMap
                glUniform1i         (locTextureMap1Unit, 0);                                        // Indicate texture unit 0 has been activated
                glBindTexture       (GL_TEXTURE_2D, [poly_entry getTextureMap1ID]);                 // Bind TextureMap to activated texture unit

                if (strcmp ([poly_entry getTextureMap1Name], dummyTexelName))                       // Check if texture map name is a placeholder
                    
                        glUniform1i (locTextureMap1Flag, true);                                     // Tell the shader the texture unit is valid
                else    glUniform1i (locTextureMap1Flag, false);                                    // Tell the shader the texture unit's invalid
                
                glActiveTexture     (GL_TEXTURE1);                                                  // Activate texture unit 1 for 2nd TextureMap
                glUniform1i         (locTextureMap2Unit, 1);                                        // Indicate texture unit 1 has been activated
                glBindTexture       (GL_TEXTURE_2D, [poly_entry getTextureMap2ID]);                 // Bind TextureMap to activated texture unit

                if (strcmp ([poly_entry getTextureMap2Name], dummyTexelName))                       // Check if texture map name is a placeholder
                    
                        glUniform1i (locTextureMap2Flag, true);                                     // Tell the shader the texture unit is valid
                else    glUniform1i (locTextureMap2Flag, false);                                    // Tell the shader the texture unit's invalid
                
                glActiveTexture		(GL_TEXTURE2);													// Activate texture unit 2 for BumpMap texture
                glUniform1i         (locBumpMapUnit, 2);                                            // Indicate texture unit 2 has been activated
                glBindTexture       (GL_TEXTURE_2D, [poly_entry getBumpMapID]);                     // Bind TextureMap to activated texture unit

                if (strcmp ([poly_entry getBumpMapName], dummyTexelName))                           // Check if texture map name is a placeholder
                    
                        glUniform1i (locBumpMapFlag, true);                                         // Tell the shader the texture unit is valid
                else    glUniform1i (locBumpMapFlag, false);                                        // Tell the shader the texture unit's invalid

#ifdef DEBUG                                                                                        // (Build Settings->MeshP2 target: Under 'Other C Flags', add '-DDEBUG=1' under 'Development')
                glValidateProgram   (ShaderProgram);                                                // Validate current context state such as sampler uniforms, current vertex array state...
                
                [self ShaderStatus  :ShaderProgram  :GL_VALIDATE_STATUS];                           // ... draw framebuffer, and so on. In short, anything that can cause draw-time errors
#endif
                glDrawElements      (GL_TRIANGLES, ([poly_entry getSurfaceIndexListCount] * 3), GL_UNSIGNED_INT, 0);    // Draw mesh with IBO->indexed to VBO->bound to VAO
            }
        }
        
        glBindVertexArray   (0);                                                                    // Unbind the VAO and its associated VBO until next Display call

        glUseProgram        (0);                                                                    // Turn off shaders in case caller uses fixed functionality OpenGL
    }
}

-(void) loadBoneTransformUniforms {                                                                 // Load glUniforms with local array elements
    
    char             name[MAX_NAMESIZE] = {0};
    NSUInteger       nAnimatedBones     = 0;
    DualQuaternion  *poseDQ;
    Quaternion      *qR;
    Quaternion      *qT;
    
    if      ([animator getAnimation  :typeArmature]) nAnimatedBones = [[[animator getAnimation  :typeArmature] getBoneAnimations] count];
    else if ([animator getAnimation  :typeAction])   nAnimatedBones = [[[animator getAnimation  :typeAction]   getBoneAnimations] count];
    else if ([animator getAnimation  :typeBVH])      nAnimatedBones = [[[animator getAnimation  :typeBVH]      getBoneAnimations] count];

    assert (nAnimatedBones < MAX_BONES);                                                            // Ensure animated bone count is valid
    
    for (int i=0; i < nAnimatedBones; i++) {                                                        // For all animated bones, Armature or Action:

        snprintf (name, sizeof(name), "BoneTransforms[%s]", [[@(i) stringValue] UTF8String]);       // Build an explicit array element name string
        
        locBoneTransforms[i] = glGetUniformLocation (ShaderProgram, name);                          // Ensure uniform hasn't been optimized out by GLSL compiler!

        if (locBoneTransforms[i] != -1) {                                                           // Is uniform location at shader's 'BoneTransforms[i]' valid?
          
            float pureDQArray[2][4];                                                                // Pure 2-column x 4-row array simulates shader's mat2x4 type

            if      (([animator getAnimation :typeArmature])) poseDQ = [ArmatureList[i] getMsAnimatedPoseTransform];   // Get (Model-space) pose for current bone
            else if ([animator getAnimation  :typeAction])    poseDQ = [ActionList[i]   getMsAnimatedPoseTransform];   //  "        "        "    "  single bone
            else if ([animator getAnimation  :typeBVH])       poseDQ = [BVHList[i]      getMsAnimatedPoseTransform];   //  "        "        "    "  BVH 'joint'
            
            if (!poseDQ) poseDQ = [DualQuaternion new];                                             // Bypass invalid bone (e.g. lacking MsAnimatedPoseTransform)
            
            qR = [DualQuaternion getReal  :poseDQ];                                                 // Get pose rotation    (i.e. 'Real' part of dual quaternion)
            qT = [DualQuaternion getDual  :poseDQ];                                                 // Get pose translation (i.e. 'Dual' part of dual quaternion)
            
            pureDQArray[0][0] = [qR getX];      pureDQArray[1][0] = [qT getX];                      // Load poseDQ into the simulated mat2x4 for upload to shader
            pureDQArray[0][1] = [qR getY];      pureDQArray[1][1] = [qT getY];
            pureDQArray[0][2] = [qR getZ];      pureDQArray[1][2] = [qT getZ];
            pureDQArray[0][3] = [qR getW];      pureDQArray[1][3] = [qT getW];
            
            glUniformMatrix2x4fv (locBoneTransforms[i],                                             // Upload to shader at verified uniform location
                                  1,                                                                // A single mat2x4 array
                                  GL_FALSE,                                                         // Column-Major order
                                  (GLfloat*)&pureDQArray);                                          // Pointer to data
        }
    }

    glUniform1ui (locAnimatedBoneCount, (GLuint)nAnimatedBones);                                    // Pass the (unsigned) bone count to the shader
}

-(void) UploadAnimation  :(Animation*)in_New_Animation  :(enum AnimationTypes)in_Type  {    // Accept new Animation object for in_Type from MeshP2 caller
    
    [animator uploadAnimation  :in_New_Animation  :in_Type];                                // Note: A 'nil' for in_Animation here stops animation
}

-(void) UploadBVHList  :(NSMutableArray*)in_newBVHList {
    
    BVHList = in_newBVHList;                                                                // Replace BVHList with externally instantiated array
    
    animator = [[Animator alloc] init  :ArmatureList  :ActionList  :BVHList];               // Replace Animator object with updated BVHList array
}

-(void) destroyVAOVBOandIBO {

    GLint   nAttrib  = 0;
    GLuint  bufferId = 0;
    
    glBindVertexArray    (meshVAO);                                 // Bind the VAO so we can delete any VBO associated within

    glGetIntegerv        (GL_MAX_VERTEX_ATTRIBS, &nAttrib);
    
    for (GLint i=0; i < nAttrib; ++i) {
        
        GLint isEnabled = GL_FALSE;
        
        glGetVertexAttribiv (i, GL_VERTEX_ATTRIB_ARRAY_ENABLED, &isEnabled);
        
        if (isEnabled) {
            
            bufferId = 0;
            
            glGetVertexAttribiv (i, GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING, (GLint*)&bufferId);  // Get the VBO for the attibute
            
            if (bufferId) { glDeleteBuffers (1, &bufferId); }       // Delete any VBO attribute set in the VAO
        }
    }
    
    glBindBuffer         (GL_ARRAY_BUFFER, 0);                      // Unbind Vertex Buffer Object (VBO)
    
    glGetIntegerv        (GL_ELEMENT_ARRAY_BUFFER_BINDING, (GLint*)&bufferId);  // Get IBO from the VAO

    if (bufferId)        { glDeleteBuffers (1, &bufferId); }        // Delete any IBO in the VAO

    glBindBuffer         (GL_ELEMENT_ARRAY_BUFFER, 0);              // Unbind the IBO

    glBindVertexArray    (0);                                       // Unbind the VAO
    
    glDeleteVertexArrays (1, &meshVAO);                             // Delete the VAO
}

-(void) dealloc {
    
    [self destroyVAOVBOandIBO];                                     // Delete Vertex and Index Buffer Objects, and Vertex Array Object

    if  (vboBuffer != nil) free (vboBuffer);                        // Free VBO buffer contents (IBOs are freed by POLYGONLIST dealloc)
    
    Scene           = nil;                                          // Destruct Scene and subordinate public object arrays
    
    VertexList      = nil;
    PolygonList     = nil;
    SurfaceList     = nil;
    ArmatureList    = nil;
    ActionList      = nil;
    BVHList         = nil;

    animator        = nil;
}

@end
