#import "OpenGL_Routines.h"

@implementation MeshP2 (OpenGL_Routines)

-(void) Convert_to_OpenGL {
/*
	Migrate internal P2* arrays built during .p2 file parsing to public MeshP2 data structures: VertexList, PolygonList, SurfaceList, ArmatureList, ActionList etc.
 
	While BVHList is *animated* by MeshP2, it's created externally by a BVH parser (e.g. DisplayP2) prior to being uploaded to MeshP2 with a call to UploadBVHList.
*/
	VertexList		= [NSMutableArray new];
	PolygonList     = [NSMutableArray new];
	SurfaceList		= [NSMutableArray new];
	ArmatureList	= [NSMutableArray new];
	ActionList      = [NSMutableArray new];
	BVHList			= nil;																	// (Instantiated externally e.g. DisplayP2->parseBVHByFileName)

    GLfloat	ambient[]	 = {0.2, 0.2, 0.2, 1.0};											// Some default material properties
    GLfloat	diffuse[]	 = {0.2, 0.2, 0.2, 1.0};
    GLfloat	specular[]	 = {0.2, 0.2, 0.2, 1.0};

	GLfloat shininess	 = 0.0;
    GLfloat	alpha		 = 1.0;

    for (P2Material *mat in [Scene getMaterial]) {											// For all materials in the scene:
		
		alpha		= 1.0 - (*[mat getTransparency] / 100);									// Invert '% transparency' to an OpenGL 'alpha opacity'

		shininess	= (*[mat getShininess]) * 128 / 100;									// Convert '% shininess' to a range from [0,128]
		
		ambient[0]	= [[mat getAmbient] getRed];											// Migrate remaining P2 material properties
		ambient[1]	= [[mat getAmbient] getGreen];											// (See http://www.sjbaker.org/steve/omniv/opengl_lighting.html)
		ambient[2]	= [[mat getAmbient] getBlue];
		ambient[3]	= 1.0;
		
		diffuse[0]	= [[mat getDiffuse] getRed];
		diffuse[1]	= [[mat getDiffuse] getGreen];
		diffuse[2]	= [[mat getDiffuse] getBlue];
		diffuse[3]	= alpha;

		specular[0]	= [[mat getSpecular] getRed];
		specular[1]	= [[mat getSpecular] getGreen];
		specular[2]	= [[mat getSpecular] getBlue];
		specular[3]	= 1.0;

		[self AddMaterial  :ambient  :diffuse  :specular  :shininess  :alpha				// Populate PolygonList with POLYGONLIST objects, indexed by material name
						   :[mat  getMatName]
						   :[[mat getTextureMap1] getTextureName]
						   :[[mat getTextureMap2] getTextureName]
						   :[[mat getBumpMap] getTextureName]
						   :*[mat getDoubleSided]];
	}
/*
	Populate initial VertexList vertices (including any UV maps) from internal source data...
*/
    unsigned int  i, j, k, numMapList, numMeshObjVertList, vertexOffset = 0;                // Initialize indices, counters, steppers

	for (P2MeshObj *p2MeshObj in [Scene getMeshObjList]) {									// For every P2MeshObj in the Scene:

		numMeshObjVertList = (int)[[p2MeshObj getVertList] count];							// Get the count of vertices in the meshObj...
        
		numMapList = (int)[[p2MeshObj getMapList] count];									// ... and the count of texture mapped vertices

		for (i=0; i < numMeshObjVertList; i++) {                                      		// For each vertex in this mesh object:
			
			[SurfaceList addObject  :[NSMutableArray new]];									// Add placeholder for the list of SURFACEs at the vertex
			
			P2Vertex *vertex_entry = [p2MeshObj getVertList][i];							// Get the current vertex at the P2MeshObj vertList array
            
			VERTEX *vertex = [VERTEX new];													// Consolidate vertex & texture data into a VERTEX object

			[vertex setX  :[vertex_entry getX]];
			[vertex setY  :[vertex_entry getY]];
			[vertex setZ  :[vertex_entry getZ]];
            
			if	((numMapList > 0) && (i < numMapList)) {									// If there is any per-vertex UV texture coordinate,
																							// it's available in the p2MeshObj->mapList array, which
				assert (numMapList == numMeshObjVertList);									// is assumed to be the same size as p2MeshObj->vertList
				
				P2Map *map_entry = [p2MeshObj getMapList][i];								// Get the UV map array entry at the corresponding vertex
				
				[vertex setU  :[map_entry getU]];											// Extract the UV texture coordinates for the new VERTEX
				[vertex setV  :[map_entry getV]];
			}
			
			if ([vertex getX] > max_X) max_X = [vertex getX];								// Remember any maximum values discovered for X, Y and Z
			if ([vertex getY] > max_Y) max_Y = [vertex getY];
			if ([vertex getZ] > max_Z) max_Z = [vertex getZ];

			[VertexList addObject  :vertex];												// Add the new vertex onto the grand unified VertexList
        }
/*
	Further populate VertexList entries with vertex flags from internal P2MeshObj->vertFlagList (Blender imported source data).
	 
	Using our Blender export plug-in (io_scene_p2), we extend the obsolete 3DS format to support multiple UV coords per vertex.
	This is done by first splitting each of Blender's original vertices; once for each unique UV coordinate sharing the vertex.
	Then, the vertex flag from the original vertex is cloned into each split vertex's flag to ensure animations work correctly.
*/
		NSMutableArray *flags = [p2MeshObj getVertFlagList];								// For this P2MeshObj, get vertex flag array

        for (j=0; j < [flags count]; j++) {                              		    		// For any P2VertexFlag type objects:
            
			uint32 longFlag  = [flags[j] getFlag];											// Get the 32-bit encoded vertex flag
			
            uint16 vertex 	 = ((longFlag & 0xFFFF0000) >> 16);                         	// Extract the target Vertex Index
			uint16 shortFlag =  (longFlag & 0x0000FFFF);									// Extract the 16-bit flag
            
            [VertexList[vertex] addFlag  :shortFlag];                   					// Add flag to target VertexList entry
		}
/*
	Populate PolygonList materials and linked surfaces from internal P2MeshObj->matList (Blender imported source data)...
*/
		TRIANGLE *triangle = [TRIANGLE new];												// Instantiate a scratch triangle object

		for (P2MaterialFacesList *matFaceList_entry in [p2MeshObj getMatList]) {			// For all material-to-faceList bindings...

			for	(k=0; k < [[matFaceList_entry getFaces] count]; k++) {						// ... and for all faces in this binding...

				int face = [[matFaceList_entry getFaces][k] intValue];						// ... read the faceList array index (e.g. faces[0] holds 23)
                
				P2Face *facelist_entry = [p2MeshObj getFaceList][face];						// De-reference that faceList entry (e.g. faceList[23])
                
				[triangle setP0  :([facelist_entry getP0] + vertexOffset)];					// ... and use the Points of that face...
				[triangle setP1  :([facelist_entry getP1] + vertexOffset)];					// ... to populate the scratch triangle object
				[triangle setP2  :([facelist_entry getP2] + vertexOffset)];					// (Again, Points are indices to VertexList entries)
                
				[self AddPolygon  :triangle  :[matFaceList_entry getName]];					// Create a SURFACE from this triangle by defining surface normals,
            }                                                                               // creating a 2nd side (if needed) and adding the new surface(s) to the
        }                                                                                   // PolygonList[material].SurfaceIndexList array of material bound surfaces
/*
	Populate ArmatureList BONE objects from internal P2MeshObj->armatureList (Blender imported source data)...
*/
        NSMutableArray *bones = [p2MeshObj getArmatureList];                           		// Get source P2MeshObj's armatureList array

        ArmatureList = [[NSMutableArray alloc] initWithCapacity  :[bones count]];        	// Initialize output target array of BONE objects

		for (P2ArmatureBone *sourceBone in bones) {                                     	// For any source bones (P2ArmatureBone objects):
			
			BONE *targetBONE = [[BONE alloc] init  :[sourceBone getIndex]                   // Initialize an output BONE in ArmatureList array
												   :[sourceBone getName]
												   :[sourceBone getParentIndex]];
			
			[targetBONE    setChildIndices  :[sourceBone getChildIndices]];              	// Copy source P2ArmatureBone Child Indices to target BONE entry
			[targetBONE	     setAnimations	:[sourceBone getAnimations]];					//  " 	   " 		 "		  Animations 	"	 " 	   "	 "
			[targetBONE setAnimationLength  :[sourceBone getAnimationLength]];				//  " 	   " 		 "	Animation Length 	"	 " 	   "	 "
			
			[targetBONE 	   setLocation  :[sourceBone getLocation]];               		// Copy    "     	 "   	  Bone Location to target BONE entry
			[targetBONE 	   setRotation  :[sourceBone getRotation]];               		//  "      "     	 "   	   "   Rotation "    "     "     "
			[targetBONE    		  setScale  :[sourceBone getScale]];                		//  "      "     	 "   	   "   Scale 	"    "     "     "
			
			[targetBONE setLsBindTransform  :[[DualQuaternion alloc] initFromMatrix  :[sourceBone getLsBindTransform]]];	// Save as DQ in target BONE entry

			NSMutableArray *boneAnimations = [sourceBone getAnimations];					// De-reference the source P2ArmatureBone's animation array entry
            
            [targetBONE      setAnimations  :boneAnimations];               				// Copy animations to the target BONE's animations array entry

			[ArmatureList insertObject  :targetBONE  atIndex:[sourceBone getIndex]];     	// Insert target BONE in ArmatureList at source bone's index
			
			if (P2_Logging_Enabled) {
				
				for (int frame = 0; frame < MAX_DEBUG_FRAMES; frame++) {                	// Readout debug values for a few joint frames
					
					Matrix4f *pose 		  = [boneAnimations objectAtIndex  :frame];			// Get the pose transform at the current frame
					
					glm::mat4 glmMat4Pose = glm::make_mat4 ([pose getPtr]);					// Convert to glm::mat4 format for display
					
					glm::quat glmQuat     = glm::quat_cast (glmMat4Pose);                 	// Get local-space rotation as a glm::quat
					
					glm::vec3 vec3Eulers  = glm::eulerAngles (glmQuat);                 	// Decompose the individual euler rotation angles
					
					NSString *strMat4Pose = [NSString stringWithUTF8String  :glm::to_string (glmMat4Pose).c_str()]; // Pose transform as matrix string
					NSString *strQuatPose = [NSString stringWithUTF8String  :glm::to_string (glmQuat).c_str()];     // Pose rotation quaternion string
					
					NSString *strFormat   = [NSString stringWithUTF8String  :"MeshP2->Convert_to_OpenGL\t(\"%@\")\tJoint:%d (\"%@\")\tFrame:%d\n"
											 "\tRowMajorMat4:\t%@\n\tQuaternion:\t\t%@\n"
											 "\tEulerXYZ:\t\t(%f\u00b0, %f\u00b0, %f\u00b0)\n"
											 "\tPosition:\t\t(%f, %f, %f)\n"];
					
					[MeshP2 P2Log  :P2_Logging_Enabled  :strFormat, [MeshFileName lastPathComponent],
											[sourceBone getIndex], [sourceBone getName], frame, strMat4Pose, strQuatPose,
											glm::degrees (vec3Eulers.x), glm::degrees (vec3Eulers.y), glm::degrees (vec3Eulers.z),
											glmMat4Pose[3].x, glmMat4Pose[3].y, glmMat4Pose[3].z];
				}
			}
		}
/*
	Populate public ActionList keyframes from internal P2MeshObj->actionList keyframes (Blender imported source data)...

	Note: An ACTION is processed as a parentless/childless Armature BONE, allowing reuse of the bone animation shader code as much as possible.
*/
		NSMutableArray *actions = [p2MeshObj getActionList];                           		// Get source array of P2MeshObj imported P2Actions

        ActionList = [[NSMutableArray alloc] initWithCapacity  :[actions count]];         	// Initialize output target array of ACTION objects
		
		assert ([actions count] == 1);														// (Currently, only a *single* action is supported)
		
		for (P2Action *sourceP2Action in actions) {                                     	// For any source P2Action objects in the P2MeshObj

			Matrix4f *matTransform = [[Matrix4f new] cmMatrixFromBytes  :[p2MeshObj getTransformMatrix]]; // Convert bind transform to Matrix4f

			[sourceP2Action setMsBindTransform  :matTransform];								// Set as (model-space) source bind transform
			
			[sourceP2Action synthesizeActionAnimations];                                	// Synthesize P2Action's psuedo-armature animations

			ACTION *targetACTION = [[ACTION alloc] init	 :[sourceP2Action getNodeID]		// Initialize an output ACTION in the ActionList array
														 :[sourceP2Action getParentNodeID]
														 :[[NSString alloc] initWithCString  :[sourceP2Action getName] encoding:NSUTF8StringEncoding]];

			[targetACTION setAnimations  	  :[sourceP2Action getAnimations]];				// Copy source P2Action animations to target ACTION object
			[targetACTION setAnimationLength  :[sourceP2Action getAnimationLength]];		// Copy source P2Action animation length to target ACTION object
			
			[targetACTION setMsBindTransform  :[[DualQuaternion alloc] initFromMatrix  :[sourceP2Action getMsBindTransform]]];  // Convert Matrix4f to DualQuaternion

            [ActionList addObject  :targetACTION];   										// Add ACTION object to the end of target ActionList array, i.e. *NOT* by ID
		}

		vertexOffset += [[p2MeshObj getVertList] count];									// Advance the VertexList stepper along to the next P2MeshObj's first vertex
    }
/*
	BVHList array? It's added to this MeshP2 object by a call to UploadBVHList from, say, DisplayP2->parseBVHByFileName, with a BVHList created during BVH file parsing.

	Subsequently, this BVHList of BONE objects is very similar to ArmatureList objects, which allows the reuse of existing bone animation shader code as much as possible.
*/
    [self LinkVerticesToPolygons];                                                          // Prepare to calculate vertex normals, as compared to the surface normals we have
	
    [self CalculateVertexNormalsAndTangents];                                               // Compute vertex normals and tangents, roughly averaged among surfaces sharing a vertex

	[self SetupVAOVBOandIBO];																// Create a VAO and VBO for all vertex attributes, plus an IBO (per material) indexing the VBO

	[self SetupShaders];																	// Create vertex and fragment shaders to handle per-pixel lighting, texturing and bump mapping
}

-(void) AddMaterial  :(GLfloat*)ambient		:(GLfloat*)diffuse		:(GLfloat*)specular  	:(GLfloat)shininess  :(GLfloat)alpha
					 :(char*)materialName	:(char*)textureMap1Name	:(char*)textureMap2Name	:(char*)bumpmapName  :(BOOL)doubleSided {
	
	bool	 bindFlag;
	
	POLYGONLIST *poly_entry	= [POLYGONLIST new];
	
	[poly_entry setName  	  :materialName];
    [poly_entry setAmbient	  :ambient];
    [poly_entry setDiffuse	  :diffuse];
    [poly_entry setSpecular	  :specular];
	[poly_entry setShininess  :shininess];
    [poly_entry setAlpha	  :alpha];
	
	if	(force2Sided) [poly_entry setDoubleSided  :true];									// Propagate flag which forces 2-Sided Materials
	else			  [poly_entry setDoubleSided  :doubleSided];							// ... otherwise, use what's in the mesh file
	
    if  (![poly_entry getTextureMap1ID]) [poly_entry genTextureMap1ID];                     // Generate Texture objectIDs ...
    if  (![poly_entry getTextureMap2ID]) [poly_entry genTextureMap2ID];                     // ... Note that we can't use ...
    if  (![poly_entry getBumpMapID])     [poly_entry genBumpMapID];                         // ... glIsTexture on unbound IDs

	if (strlen (textureMap1Name)) {															// If there's a 1st texture file

		bindFlag = [self LoadTextureFile  :textureMap1Name  :[poly_entry getTextureMap1ID]];// Load it, reporting any errors
		
		[poly_entry setTextureMap1Name  :textureMap1Name];									// Set 1st material texture name

	} else {
				
		bindFlag = [self Load1x1PixelTexture  :[poly_entry getTextureMap1ID]];				// Otherwise, load 1x1 default texel for sampler2D
																							// See https://stackoverflow.com/a/42239068/848364
		[poly_entry setTextureMap1Name  :dummyTexelName];									// Set placeholder texture name
	}
	
	[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->AddMaterial:'%s' Texture_ID:%u, TextureMap1Name:'%s', Bound:%@",
		   materialName, [poly_entry getTextureMap1ID], [poly_entry getTextureMap1Name], (bindFlag ? @"Yes" : @"*** NO ***")];

	if (strlen (textureMap2Name)) {															// If there's a 2nd texture file

		bindFlag = [self LoadTextureFile  :textureMap2Name  :[poly_entry getTextureMap2ID]];// Load it, reporting any errors
		
		[poly_entry setTextureMap2Name  :textureMap2Name];									// Set 2nd material texture name

	} else {
		
		bindFlag = [self Load1x1PixelTexture  :[poly_entry getTextureMap2ID]];				// Load 1x1 texel for sampler2D
		
		[poly_entry setTextureMap2Name  :dummyTexelName];									// Set placeholder texture name
	}

	[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->AddMaterial:'%s' Texture_ID:%u, TextureMap2Name:'%s', Bound:%@",
		   materialName, [poly_entry getTextureMap2ID], [poly_entry getTextureMap2Name], (bindFlag ? @"Yes" : @"*** NO ***")];

	if (strlen (bumpmapName)) {																// If there's a bump map texture

		bindFlag = [self LoadTextureFile  :bumpmapName  :[poly_entry getBumpMapID]];		// Load it, reporting any errors

		[poly_entry setBumpMapName  :bumpmapName];											// Set the bump map texture name
		
	} else {
		
		bindFlag = [self Load1x1PixelTexture  :[poly_entry getBumpMapID]];					// Load 1x1 texel for sampler2D

		[poly_entry setBumpMapName  :dummyTexelName];										// Set placeholder texture name
	}

	[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->AddMaterial:'%s' BumpMap_ID:%u, BumpMapName:'%s', Bound:%@",
		   materialName, [poly_entry getBumpMapID], [poly_entry getBumpMapName], (bindFlag ? @"Yes" : @"*** NO ***")];

	[PolygonList addObject  :poly_entry];
}

-(void) AddPolygon  :(TRIANGLE*)triangle  :(char*)MaterialName {
	
    int MaterialNum	= [self FindMaterialIndex  :MaterialName];								// Find the material's PolygonList array index
	
	if (MaterialNum == -1) return;															// Bypass unlikely case of unknown material index
	
    POLYGONLIST *poly_entry	= PolygonList[MaterialNum];										// De-reference the listhead of POLYGONLIST entries
	
    SURFACE	*surfaceIndex_entry	= [poly_entry addSurfaceIndexList];							// Create a new SURFACE index object bound to this material
	
    [surfaceIndex_entry setP0  :[triangle getP0]];											// Populate the new surface's index Points...
    [surfaceIndex_entry setP1  :[triangle getP1]];
    [surfaceIndex_entry setP2  :[triangle getP2]];
	
    [self CalculateSurfaceNormal  :surfaceIndex_entry];										// ... as well as the surface's normal vector
	
    if	([poly_entry getDoubleSided]) {														// If this material is double-sided:

		SURFACE *secondSurfaceIndex_entry = [poly_entry addSurfaceIndexList];				// Create a new surface bound to the material
		
		[secondSurfaceIndex_entry setP0  :[triangle getP0]];								// Note: this 2nd surface is wound in reverse
		[secondSurfaceIndex_entry setP1  :[triangle getP2]];
		[secondSurfaceIndex_entry setP2  :[triangle getP1]];
		
		[self CalculateSurfaceNormal  :secondSurfaceIndex_entry];							// Calculate the 2nd surface's normal vector
	}
}

-(int)  FindMaterialIndex  :(char*)materialName {
	
    for (int n=0; n < [PolygonList count]; n++) {

		POLYGONLIST *poly = PolygonList[n];
		
		if (strcmp (materialName, [poly getName]) == 0) return n;
	}
	
    int choice =  (int)NSRunAlertPanel  (@"MeshP2->FindMaterialIndex reports:",
										 @"Failed to locate material '%s'\nwithin PolygonList!\n\n(Quit will also copy material name to clipboard)",
										 @"Ignore", @"Quit", nil, materialName);
	
    if (choice == NSAlertDefaultReturn) return (-1);
	
	if (choice == NSAlertAlternateReturn) {
		
		NSPasteboard *pasteBoard = [NSPasteboard generalPasteboard];
		
		[pasteBoard declareTypes  :[NSArray arrayWithObjects  :NSStringPboardType, nil] owner:nil];
		[pasteBoard setString  	  :[NSString stringWithUTF8String  :materialName] forType:NSStringPboardType];
		
		ExitToShell();
	}
	
    return (-1);
}

-(bool) CalculateSurfaceNormal  :(SURFACE*)surface {										// Refer to http://www.opengl.org/wiki/Calculating_a_Surface_Normal
	
    Vector3f	*normal = [Vector3f new];

    GLfloat		Ax, Ay, Az;
    GLfloat		Bx, By, Bz;
	
    GLfloat		length = 0.0;

    VERTEX *p0 = VertexList[[surface getP0]];
    VERTEX *p1 = VertexList[[surface getP1]];
    VERTEX *p2 = VertexList[[surface getP2]];
	
    Ax = [p1 getX] - [p0 getX];
    Ay = [p1 getY] - [p0 getY];
    Az = [p1 getZ] - [p0 getZ];
	
    Bx = [p2 getX] - [p0 getX];
    By = [p2 getY] - [p0 getY];
    Bz = [p2 getZ] - [p0 getZ];
	
	[normal setX  :(Ay*Bz - Az*By)];														// Find the cross product
	[normal setY  :(Az*Bx - Ax*Bz)];
	[normal setZ  :(Ax*By - Ay*Bx)];
	
    length = ([normal getX] * [normal getX]) +												// Make the normal a unit vector
			 ([normal getY] * [normal getY]) +
			 ([normal getZ] * [normal getZ]);
	
	if (!length) return false;																// Ignore degenerate polygons - most graphics cards won't draw them anyway
	
    length = (GLfloat) 1.0 / (GLfloat) sqrt (length);
	
	[normal setX  :([normal getX] * length)];
	[normal setY  :([normal getY] * length)];
	[normal setZ  :([normal getZ] * length)];

	[surface setNX  :[normal getX]];
    [surface setNY  :[normal getY]];
    [surface setNZ  :[normal getZ]];
	
    return true;
}

-(void) LinkVerticesToPolygons {
/*
	Prior to calculating vertex normals and tangents, we need to link all vertices to the polygons that reference them.
 
	For each material in the PolygonList array:
	 For each surface bound to the material:
	  For each point of the surface:

	   - Get the vertex number of the Point; suitable for referencing into an IBO buffer
	   - Index into the SurfaceList for the listhead of surfaces at/sharing the vertex
	   - Add an entry for the surface to the list of surfaces referencing the vertex
 
	Subsequently, CalculateVertexNormalsAndTangents will use the SurfaceList to compute vertex normals and tangents.
*/
    for (POLYGONLIST *poly_entry in PolygonList) {

		for (SURFACE *surface_entry in [poly_entry getSurfaceIndexList]) {
			
			[SurfaceList[[surface_entry getP0]] addObject  :surface_entry];
			[SurfaceList[[surface_entry getP1]] addObject  :surface_entry];
			[SurfaceList[[surface_entry getP2]] addObject  :surface_entry];
		}
	}
}

-(void) CalculateVertexNormalsAndTangents {
/*
	Compute vertex normal and tangent vectors, averaged across any adjacent surfaces sharing the vertex, for use by the fragment shader.
 
	Vector normalization is deferred to the vertex shader, while the fragment shader performs Gram-Schmidt orthonormalization and per-pixel lighting.
*/
    SURFACE	*firstSurface, *currentSurface;															// Pointers to the first surface at this vertex and the current surface if the vertex is shared

    for (int i = 0; i < [VertexList count]; i++) {													// For every vertex in the mesh (*including* any new 'split' vertices added within this loop):

		VERTEX *vertex_entry = VertexList[i];														// Get a pointer to the current vertex entry
		
		[vertex_entry setNX  :0.0f];  [vertex_entry setNY  :0.0f];  [vertex_entry setNZ  :0.0f];	// Initialize vertex normal vector
		
		[vertex_entry setTX  :0.0f];  [vertex_entry setTY  :0.0f];  [vertex_entry setTZ  :0.0f];	// Initialize vertex tangent vector
		
		NSMutableArray *surfaces = SurfaceList[i];													// Get the list of surfaces at this vertex

		NSUInteger numSurfaces = [surfaces count];													// Get the number of surfaces at this vertex
		
		if (numSurfaces == 0) continue;																// Ignore isolated vertices (points & lines)

		firstSurface = surfaces[0];																	// Note the first surface for comparison

		for (int surface = 0; surface < numSurfaces; surface++) {									// For each surface referencing this vertex:

			currentSurface = surfaces[surface];														// Get a pointer to the current surface object

			Vector3f *vertexTangent = [self Tangent  :currentSurface  :i];							// Compute the vertex tangent vector for bump mapping

			if  (numSurfaces == 1) {																// For a vertex having only one surface referencing it:

				[vertex_entry setNX  :[currentSurface getNX]];										// Propagate the current surface normal as vertex normals
				[vertex_entry setNY  :[currentSurface getNY]];
				[vertex_entry setNZ  :[currentSurface getNZ]];
				
				[vertex_entry setTX  :[vertexTangent getX]];										// Record the computed vertex tangent vector
				[vertex_entry setTY  :[vertexTangent getY]];
				[vertex_entry setTZ  :[vertexTangent getZ]];

			} else {																				// Otherwise, there are at least 2 surfaces referencing this vertex. In this case, determine:
				
				if	([self Cos_angle  :firstSurface  :currentSurface] <= cosineThreshold) {			// Do we need to split this vertex? For an example threshold angle of 90°, the cosine(90°) is 0.0.
																									// The dot product of two surface normals whose angle between them >= 90° will be a cosine <= 0.0,
																									// ∴ if the angle between the first and current surface normals is >= 90°, then split this vertex:
					
					VERTEX	*new_vertex = [[VERTEX alloc] init	:[vertex_entry getX]	:[vertex_entry getY]	:[vertex_entry getZ]	// position
																:0.0					:0.0					:0.0					// normal
																:0.0					:0.0					:0.0					// tangent
																:[vertex_entry getU]	:[vertex_entry getV]							// texcoord
															    :[vertex_entry getFlags]];												// vertex flags
					
					[VertexList addObject  :new_vertex];											// Add the cloned vertex to VertexList (normals & tangents are added when outer for-loop hits new vertex)
					
					int lastVertex = (int)[VertexList count] - 1;									// Get an int index to the vertex we just added (saves some dereferencing)
					
					if ([currentSurface getP0] == i) [currentSurface setP0  :(lastVertex)];			// Remap whichever surface point needs to change to the newly added vertex
					if ([currentSurface getP1] == i) [currentSurface setP1  :(lastVertex)];
					if ([currentSurface getP2] == i) [currentSurface setP2  :(lastVertex)];
					
					[SurfaceList addObject  :[NSMutableArray new]];									// Add an empty surface array listhead for the new vertex onto SurfaceList
					
					[[SurfaceList lastObject] addObject  :currentSurface];							// Add the re-mapped surface to the empty surface array for subsequent outer loop processing
					
				} else {																			// If not splitting, accumulate the vertex normal & tangent vectors across all unsplit surfaces

					[vertex_entry setNX  :([vertex_entry getNX] + [currentSurface getNX])];			// These accumulated vectors will be pseudo-averaged when the vertex shader performs normalization
					[vertex_entry setNY  :([vertex_entry getNY] + [currentSurface getNY])];
					[vertex_entry setNZ  :([vertex_entry getNZ] + [currentSurface getNZ])];
					
					[vertex_entry setTX  :([vertex_entry getTX] + ([vertexTangent getX]))];
					[vertex_entry setTY  :([vertex_entry getTY] + ([vertexTangent getY]))];
					[vertex_entry setTZ  :([vertex_entry getTZ] + ([vertexTangent getZ]))];
				}
			}
		}
	}
}

-(void) SetupVAOVBOandIBO {
/*
	First, build a monolithic client-side VBO buffer containing interleaved vertex data from all materials, combined as:
 
	11 GLfloats per VERTEX array element (3 Vertex Coords, 3 Normals, 3 Tangents and 2 Texture Coords)
	+4 GLuints  for Bone Indices of bones deforming this vertex (for Vertex Weighting/Skinning)
	+4 GLfloats for Bone Weights of bones deforming this vertex (sorted and normalized)
	--
	19 4-byte values, totaling 76 attribute bytes per vertex; defined as symbol (below)
*/

#define ATTRIBUTE_BYTES_PER_VERTEX (19 * 4)                  										// Attribute Bytes per VertexList element (avoids vboSize/vboBuffer/glVertexAttribPointer typos)
	
    uint    vertexCount = (uint)[VertexList count];													// Unlike CalculateVertexNormalsAndTangents, we can safely de-reference [VertexList count] here
    
    vboSize = vertexCount * (ATTRIBUTE_BYTES_PER_VERTEX);											// Per the above breakdown: 19 x 4-bytes each, for a total of 76 attribute bytes per vertex
    
	do {																							// Collapse the VertexList array to interleaved VBO format for uploading to the GPU

		if	(vboBuffer != nil) free (vboBuffer);
		
		if	((vboBuffer = (GLfloat*)calloc (vertexCount, ATTRIBUTE_BYTES_PER_VERTEX)) != nil) {		// (this local client-side VBO buffer gets freed in MeshP2->dealloc)

			GLfloat *vboBufferPtr = vboBuffer;                                                  	// An incremented pointer to the current vertex's vboBuffer entry
			
			for	(int i=0; i < vertexCount; i++) {													// Copy the vertex attributes into the client-side VBO buffer

				VERTEX  *vertex_entry = VertexList[i];

				*vboBufferPtr++ = [vertex_entry getX];                                          	// Vertex position coordinates (3 floats)
				*vboBufferPtr++ = [vertex_entry getY];
				*vboBufferPtr++ = [vertex_entry getZ];

				*vboBufferPtr++ = [vertex_entry getNX];					                        	// Vertex normal vector (3 floats)
				*vboBufferPtr++ = [vertex_entry getNY];
				*vboBufferPtr++ = [vertex_entry getNZ];

				*vboBufferPtr++ = [vertex_entry getTX];					                        	// Vertex tangent vector (3 floats)
				*vboBufferPtr++ = [vertex_entry getTY];
				*vboBufferPtr++ = [vertex_entry getTZ];

				*vboBufferPtr++ = [vertex_entry getU];					                        	// Vertex texcoords (2 floats)
				*vboBufferPtr++ = [vertex_entry getV];

				[vertex_entry BuildBoneIndicesAndWeightsArray];										// Build sorted/normalized Bone Indices and Weights array
				
				NSArray *vertex_bones = [vertex_entry getBoneIndicesAndWeights];

				*vboBufferPtr++ = [vertex_bones[0][0] unsignedIntValue];      						// 1st Bone Index (or MAX_BONES as an out-of-range indicator)
				*vboBufferPtr++ = [vertex_bones[1][0] unsignedIntValue];      						// 2nd  "     "    "  "
				*vboBufferPtr++ = [vertex_bones[2][0] unsignedIntValue];      						// 3rd  "     "    "  "
				*vboBufferPtr++ = [vertex_bones[3][0] unsignedIntValue];      						// 4th  "     "    "  "

				*vboBufferPtr++ = [vertex_bones[0][1] floatValue];      							// 1st Bone Weight (or 0.0)
				*vboBufferPtr++ = [vertex_bones[1][1] floatValue];      							// 2nd  "     "     "   "
				*vboBufferPtr++ = [vertex_bones[2][1] floatValue];      							// 3rd  "     "     "   "
				*vboBufferPtr++ = [vertex_bones[3][1] floatValue];      							// 4th  "     "     "   "
			}

		} else {
			
			int choice = (int) NSRunAlertPanel  (@"MeshP2->SetupVAOVBOandIBOs reports:",
												 @"Failed to calloc %d bytes of local VBO memory!",
												 @"Quit", @"Retry", nil, vboSize);
			
			if (choice == NSAlertDefaultReturn) ExitToShell();
			if (choice == NSAlertAlternateReturn) continue;
		}
		
	} while (vboBuffer == nil);															//FIXME: Shouldn't loop forever if calloc fails

/*
	Next, create a Vertex Array Object (VAO), with a Vertex Buffer Object (VBO) server-side (i.e. on the GPU) to receive the client-side vboBuffer contents
*/
	if (meshVAO)			[self destroyVAOVBOandIBO];									// Remove any residual VAO and VBO, for example from parsing a previous mesh
							 
	glGenVertexArrays		(1, &meshVAO);												// Generate a unique ID for the single VAO (Vertex Array Object) of the mesh
	glGenBuffers			(1, &meshVBO);												// Generate a unique ID for the single VBO (Vertex Buffer Object) of the mesh

	glBindVertexArray   	(meshVAO);                                            		// Bind to VAO so that the following state changes are captured in the VAO

    glBindBuffer			(GL_ARRAY_BUFFER, meshVBO);									// Bind the VBO to GPU memory, which also 'associates' the VBO to the VAO
	
    glBufferData			(GL_ARRAY_BUFFER, vboSize, vboBuffer, GL_STATIC_DRAW);		// Upload vboBuffer to the VBO, ∴ the GPU; it's freed in MeshP2->dealloc

	glVertexAttribPointer	(0, 3, GL_FLOAT, GL_FALSE, ATTRIBUTE_BYTES_PER_VERTEX, (const GLvoid*)0);	// Attribute 0: Vertex Position 	(3 floats)
	glVertexAttribPointer	(1, 3, GL_FLOAT, GL_FALSE, ATTRIBUTE_BYTES_PER_VERTEX, (const GLvoid*)12);	//     "     1:    "   Normal   	(3 floats)
	glVertexAttribPointer	(2, 3, GL_FLOAT, GL_FALSE, ATTRIBUTE_BYTES_PER_VERTEX, (const GLvoid*)24);	//     "     2:    "   Tangent  	(3 floats)
	glVertexAttribPointer	(3, 2, GL_FLOAT, GL_FALSE, ATTRIBUTE_BYTES_PER_VERTEX, (const GLvoid*)36);	//     "     3:    "   TexCoord 	(2 floats)
	glVertexAttribPointer	(4, 4, GL_FLOAT, GL_FALSE, ATTRIBUTE_BYTES_PER_VERTEX, (const GLvoid*)44);	//     "     4:    "   Bone Index  	(4 floats)
	glVertexAttribPointer	(5, 4, GL_FLOAT, GL_FALSE, ATTRIBUTE_BYTES_PER_VERTEX, (const GLvoid*)60);	//     "     5:    "   Bone Weight 	(4 floats)
	
    glEnableVertexAttribArray (0);                                                    	// Enable attribute 0, bound to 'Position' attribute in vertex shader
    glEnableVertexAttribArray (1);                                                   	//     "      "     1,   "   "  'Normal'		"	 "	   "	 "
    glEnableVertexAttribArray (2);                                                    	//     "      "     2,   "   "  'Tangent'		"	 "	   "	 "
    glEnableVertexAttribArray (3);                                                   	//     "      "     3,   "   "  'TexCoord'		"	 "	   "	 "
    glEnableVertexAttribArray (4);                                                   	//     "      "     4,   "   "  'BoneIndices'	"	 "	   "	 "
    glEnableVertexAttribArray (5);                                                   	//     "      "     5,   "   "  'BoneWeights'	"	 "	   "	 "
	
/*
	Lastly, for each material, create a unique Index Buffer Object (IBO) to index the VBO entries for the material's interleaved vertices, normals, texcoords, tangents, bone indices and weights
*/

	for	(POLYGONLIST *poly_entry in PolygonList) {										// For each material in the PolygonList array:

        int surfaceCount = [poly_entry getSurfaceIndexListCount];						// Get the number of surfaces bound to this material
        
		if (surfaceCount == 0) continue;												// Skip this material if it has no surfaces bound to it

		[poly_entry setIBOSize  :(surfaceCount * 3 * sizeof(GLuint))];					// 3 GLuint Points per SurfaceIndexList (i.e. SURFACE) array element, each point indexes into the VBO

		GLuint *iboBuffer = [poly_entry getIBOBuffer];									// Get the pointer to this material's IBO buffer
		
		do {

			if	(iboBuffer != nil) free (iboBuffer);									// Free any previous iboBuffer contents
			
			if	((iboBuffer = (GLuint*)malloc ([poly_entry getIBOSize])) == nil) {		// (iboBuffer will be freed during this PolygonList entry's dealloc method)

				int choice =  (int) NSRunAlertPanel  (@"MeshP2->SetupVAOVBOandIBOs reports:",
													  @"Failed to malloc %d bytes of local IBO memory!",
													  @"Quit", @"Retry", nil, [poly_entry getIBOSize]);
				
				if (choice == NSAlertDefaultReturn) ExitToShell();
				if (choice == NSAlertAlternateReturn) continue;
			}
			
		} while (iboBuffer == nil);														//FIXME: Shouldn't loop forever if malloc fails
		
        [poly_entry setIBOBuffer  :iboBuffer];											// Save the local IBO buffer pointer to the PolygonList entry
        
		int k = 0;

		for (SURFACE *surfaceIndex_entry in [poly_entry getSurfaceIndexList]) {			// For each surface in this material, load index values (i.e. surface Points) into the IBO buffer

			iboBuffer[k++]	= [surfaceIndex_entry getP0];
			iboBuffer[k++]	= [surfaceIndex_entry getP1];
			iboBuffer[k++]	= [surfaceIndex_entry getP2];
		}

		[poly_entry genMaterialIBO];													// Generate and Set the Index Buffer Object for this material

		glBindBuffer	(GL_ELEMENT_ARRAY_BUFFER, [poly_entry getMaterialIBO]);			// Bind to the IBO buffer, indexing into the VBO when drawing
		
		glBufferData	(GL_ELEMENT_ARRAY_BUFFER, [poly_entry getIBOSize], [poly_entry getIBOBuffer], GL_STATIC_DRAW);	// Upload this material's IBO
		
		glBindBuffer	(GL_ELEMENT_ARRAY_BUFFER, 0);									// Unbind the IBO from client-side until Display is called
	}
	
	glBindVertexArray	(0);                                            				// Unbind the VAO (and associated VBO) from client-side until Display is called
}

-(void) SetupShaders {
	
    NSBundle *frameworkBundle    = [NSBundle bundleForClass  :[MeshP2 class]];
    
	NSString *vertexShaderPath   = [frameworkBundle pathForResource  :@"Shader" ofType:@"vert"];	// Load Vertex shader source code
    NSString *fragmentShaderPath = [frameworkBundle pathForResource  :@"Shader" ofType:@"frag"];	// Load Fragment shader source code

    VertexShader   		= [self CompileShaderType  :GL_VERTEX_SHADER    :vertexShaderPath];
    FragmentShader 		= [self CompileShaderType  :GL_FRAGMENT_SHADER  :fragmentShaderPath];

	[self ShaderStatus  :VertexShader    :GL_COMPILE_STATUS];
	[self ShaderStatus  :FragmentShader  :GL_COMPILE_STATUS];
	
	ShaderProgram		= glCreateProgram();											// Create the Shader program

	glAttachShader		 (ShaderProgram, VertexShader);									// Attach Vertex and Fragment Shaders
	glAttachShader		 (ShaderProgram, FragmentShader);

	glBindAttribLocation (ShaderProgram, 0, "Position");								// Bind attribute locations before linking program
	glBindAttribLocation (ShaderProgram, 1, "Normal");
	glBindAttribLocation (ShaderProgram, 2, "Tangent");
	glBindAttribLocation (ShaderProgram, 3, "TexCoord");
	glBindAttribLocation (ShaderProgram, 4, "BoneIndices");
	glBindAttribLocation (ShaderProgram, 5, "BoneWeights");

	glLinkProgram		 (ShaderProgram);												// Link the shader program to bound attribute locations
	
	[self ShaderStatus   :ShaderProgram  :GL_LINK_STATUS];
	
	glUseProgram 		 (ShaderProgram);
	
	glDeleteShader       (VertexShader);												// Delete unneeded Vertex and Fragment shader objects
	glDeleteShader       (FragmentShader);
	
	locModelViewMatrix           = glGetUniformLocation (ShaderProgram, "ModelViewMatrix");
	locProjectionMatrix          = glGetUniformLocation (ShaderProgram, "ProjectionMatrix");

	locAnimatedBoneCount   		 = glGetUniformLocation (ShaderProgram, "AnimatedBoneCount");

	locGlobalLightPosition       = glGetUniformLocation (ShaderProgram, "LightSourcePosition");
	locGlobalLightAmbient        = glGetUniformLocation (ShaderProgram, "LightSourceAmbient");
	locGlobalLightDiffuse        = glGetUniformLocation (ShaderProgram, "LightSourceDiffuse");
	locGlobalLightSpecular       = glGetUniformLocation (ShaderProgram, "LightSourceSpecular");
	
	locGlobalLightConstantAttenuation  = glGetUniformLocation (ShaderProgram, "LightConstantAttenuation");
	locGlobalLightLinearAttenuation    = glGetUniformLocation (ShaderProgram, "LightLinearAttenuation");
	locGlobalLightQuadraticAttenuation = glGetUniformLocation (ShaderProgram, "LightQuadraticAttenuation");
	
	locMaterialLightAmbient    	 = glGetUniformLocation (ShaderProgram, "MaterialAmbient");
	locMaterialLightDiffuse   	 = glGetUniformLocation (ShaderProgram, "MaterialDiffuse");
	locMaterialLightSpecular   	 = glGetUniformLocation (ShaderProgram, "MaterialSpecular");
	
	locMaterialShininess         = glGetUniformLocation (ShaderProgram, "MaterialShininess");
	locMaterialAlpha             = glGetUniformLocation (ShaderProgram, "MaterialAlpha");
	
	locTextureMap1Flag           = glGetUniformLocation (ShaderProgram, "TextureMap1Flag");
	locTextureMap1Unit           = glGetUniformLocation (ShaderProgram, "TextureMap1Unit");
	
	locTextureMap2Flag           = glGetUniformLocation (ShaderProgram, "TextureMap2Flag");
	locTextureMap2Unit           = glGetUniformLocation (ShaderProgram, "TextureMap2Unit");

	locBumpMapFlag           	 = glGetUniformLocation (ShaderProgram, "BumpMapFlag");
	locBumpMapUnit           	 = glGetUniformLocation (ShaderProgram, "BumpMapUnit");
	
	glUseProgram		(0);
}

-(void) ShaderStatus  :(GLuint)Object  :(GLenum)Parameter {
	
	GLchar	messages[256];
	GLint	status, logMessageLength;
	
	switch (Parameter) {
			
		case GL_COMPILE_STATUS:
			
			glGetShaderiv (Object, GL_COMPILE_STATUS, &status);
			
			if (status == GL_FALSE) {
				
				glGetShaderInfoLog (Object, sizeof(messages), 0, &messages[0]);
				
				[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ShaderStatus reports GL_COMPILE_STATUS for Shader #%d as: %s", Object, messages];
				
				glDeleteShader (Object);
				
				exit (-1);
			}
			
			break;
			
		case GL_LINK_STATUS:
			
			glGetProgramiv (Object, GL_LINK_STATUS, &status);
			
			if (status == GL_FALSE) {
				
				glGetProgramiv (Object, GL_INFO_LOG_LENGTH, &logMessageLength);
				
				if (logMessageLength == 0) {
					
					[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ShaderStatus reports GL_LINK_STATUS as GL_FALSE, but without any further message text"];
					
				} else {
					
					glGetProgramInfoLog (Object, sizeof(messages), 0, &messages[0]);
					
					[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ShaderStatus reports GL_LINK_STATUS as GL_FALSE, with message text as: %s", messages];
				}
				
				glDetachShader	(ShaderProgram, VertexShader);
				glDetachShader	(ShaderProgram, FragmentShader);
				
				glDeleteShader	(VertexShader);
				glDeleteShader	(FragmentShader);
				
				ShaderProgram	= 0;
				
				exit (-1);
			}
			
			break;
			
		case GL_VALIDATE_STATUS:
			
			glGetProgramiv (Object, GL_VALIDATE_STATUS, &status);

			if (status == GL_FALSE) {
				
				glGetProgramiv (Object, GL_INFO_LOG_LENGTH, &logMessageLength);
				
				if (logMessageLength == 0) {
					
					[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ShaderStatus reports GL_VALIDATE_STATUS as GL_FALSE, but without any further message text"];
					
				} else {
					
					glGetProgramInfoLog (Object, sizeof(messages), 0, &messages[0]);
					
					[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ShaderStatus reports GL_VALIDATE_STATUS as GL_FALSE, with message text as: %s", messages];
				}
				
				glDetachShader	(ShaderProgram, VertexShader);
				glDetachShader	(ShaderProgram, FragmentShader);
				
				glDeleteShader	(VertexShader);
				glDeleteShader	(FragmentShader);
				
				ShaderProgram	= 0;
				
				exit (-1);
			}
			
			break;
	}
}

-(GLuint) CompileShaderType  :(GLenum)type  :(NSString*)file {
	
	NSError    *error;
	char        sourceAsCString[20480];
	const char *ptr_ShaderSource = sourceAsCString;
	NSString   *sourceAsNSString = [NSString stringWithContentsOfFile  :file encoding:NSUTF8StringEncoding error:&error];

	if  (![sourceAsNSString getCString  :sourceAsCString
							   maxLength:sizeof(sourceAsCString)/sizeof(*sourceAsCString)-1
								encoding:NSUTF8StringEncoding]) {
		
		[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->compileShaderOfType sourceAsCString buffer too small!"];
		
		[NSException raise  :@"Failed to initialise OpenGL" format:@"Failed to read shader file: %@", file];
		
		return 0;
	}
	
	GLuint shader = glCreateShader (type);

	glShaderSource	(shader, 1, &ptr_ShaderSource, NULL);

	glCompileShader (shader);

	return shader;
}

@end
