#version 150                                                                        // Specify OpenGL 3.2

#define MAX_LIGHTS  1                                                               //FIXME: Support 8 available lights

uniform mat4  ModelViewMatrix;                                                      // ModelView matrix used to get eye space
uniform mat4  ProjectionMatrix;                                                     // Current Projection matrix, also passed in during each frame update

uniform vec4  LightSourcePosition;                                                  // Simulate (deprecated) 'gl_LightSource' built-in containing light position, ambient, diffuse and specular vectors
uniform vec4  LightSourceAmbient;
uniform vec4  LightSourceDiffuse;
uniform vec4  LightSourceSpecular;

uniform float LightConstantAttenuation;                                             // Continue reconstituted 'gl_LightSource' built-in, for constantAttenuation, linearAttenuation and quadraticAttenuation scalars
uniform float LightLinearAttenuation;
uniform float LightQuadraticAttenuation;

uniform vec4  MaterialAmbient;                                                      // Material lighting vectors
uniform vec4  MaterialDiffuse;
uniform vec4  MaterialSpecular;

uniform float MaterialShininess;                                                    // Material lighting scalars
uniform float MaterialAlpha;

uniform bool  TextureMap1Flag;                                                      // Flags whether TextureMap1Unit contains a valid texture unit number
uniform bool  TextureMap2Flag;                                                      // Flags whether TextureMap2Unit contains a valid texture unit number
uniform bool  BumpMapFlag;                                                          // Flags whether BumpMapUnit contains a valid texture unit number

uniform sampler2D TextureMap1Unit;                                                  // 1st TextureMap's texture unit number
uniform sampler2D TextureMap2Unit;                                                  // 2nd TextureMap's texture unit number
uniform sampler2D BumpMapUnit;                                                      // Bump map's texture unit number; different than TextureMap unit

in  vec4 position;                                                                  // Fragment position, normal, tangent and texcoord interpolated across vertices of the current primitive
in  vec3 normal;
in  vec3 tangent;
in  vec2 texcoord;
out vec4 FragColour;                                                                // Output to next stage in OpenGL pipeline

vec3 CalcBumpedNormal() {                                                           // Calculate an object space bump map normal to give the illusion of depth

    vec3 BumpedNormal;

    vec3 N  = normalize (normal);                                                   // Build the tangent->object space 'Basis' (TBN Matrix) needed for transformations

    vec3 T  = normalize (tangent);                                                  // TBN Matrix is made up of the interpolated vertex Tangent, (derived) Bi-Tangent and Normals

         T  = normalize (T - dot (T, N) * N);                                       // Perform Gram-Schmidt processing of vertex normal and tangent to ensure an orthonormal basis

    vec3 B  = cross (T, N);                                                         // Derived Bi-Tangent

    BumpedNormal  =  texture (BumpMapUnit, texcoord.xy).rgb;                        // Extract the tangent space bump map normal (a vector of 3 floats) from the bump map texcoords

    BumpedNormal  =  (BumpedNormal * 2.0) - vec3 (1.0);                             // Expand the RGB encoded floats from ([0..1]) to floats of XYZ coordinates ([-1..+1])

    return normalize (BumpedNormal * mat3 (T, B, N));                               // Transform the tangent space bump map normal vector to object space
}

void main (void) {

    vec4 ambientMaterialLightProduct  = MaterialAmbient  * LightSourceAmbient;      // Simulate (deprecated) gl_FrontMaterial * gl_LightSource[i] functionality //TODO: Extend for >1 light
    vec4 diffuseMaterialLightProduct  = MaterialDiffuse  * LightSourceDiffuse;
    vec4 specularMaterialLightProduct = MaterialSpecular * LightSourceSpecular;
    
    vec3 N;                                                                         // Normal vector
    vec3 L;                                                                         // Light direction vector
    vec3 V;                                                                         // Vector from vertex to the viewer
    vec3 H;                                                                         // Half-vector between light and viewer
    
    float attenuation  = 1.0;                                                       // Distance factor to attenuated point light source
    
    vec4 fragmentColor = vec4 (0.0, 0.0, 0.0, 1.0);                                 // Ambient colour of the scene not affected by light

    for (int i=0; i<MAX_LIGHTS; i++) {                                              // For all lights:

        if (LightSourcePosition.w == 0.0) {                                         // Determine if this is a Directional or Point light source

            L = LightSourcePosition.xyz;                                            // Position means direction from which infinite light rays are cast

            attenuation = 1.0;                                                      // Directional lights are infinitely distant, so attenuation is disabled

        } else {                                                                    // Otherwise, attenuate the Point/Spot light

            L = LightSourcePosition.xyz - position.xyz;                             // Vector from fragment to light source

            float D = length (L);                                                   // Distance to light source

            attenuation = 1.0 / (LightConstantAttenuation  +                        // Attenuate by distance
                                 LightLinearAttenuation    * D +
                                 LightQuadraticAttenuation * D * D);
        }

        N = (BumpMapFlag) ? CalcBumpedNormal() : normalize (normal);                // If bump mapping, transform bump map normal from tangent->object space; else, use interpolated vertex normal

        L = normalize (L);                                                          // Normalize the light direction vector
        
        V = -(ModelViewMatrix * position).xyz;                                      // Vector from fragment to Eye position in View space
        
        H = normalize (V + L);                                                      // Normalize the half-vector between viewer and light vectors

        float NdotL    = max (dot (N, L), 0.0);                                     // Compute the dot product of the selected normal vector and light direction vectors, or at least 0.0
        
        float NdotH    = max (dot (N, H), 0.0);                                     // And the dot product of selected normal and halfway-between-viewer-and-light vectors, also >= 0.0

        float power    = (NdotL == 0.0) ? 0.0 : pow (NdotH, MaterialShininess);     // If N∙L is 0.0, the light is 'behind' the surface and outside [-90˚, +90˚], so no specularity
                                                                                    // Otherwise, raise N∙H to the power of the material shininess to define specular highlights
        vec4  ambient  = ambientMaterialLightProduct;
        vec4  diffuse  = diffuseMaterialLightProduct  * NdotL;                      //TODO: Support material ambient/diffuse/etc multiplied by light[n].ambient/diffuse/etc
        vec4  specular = specularMaterialLightProduct * power;

        fragmentColor += ((ambient + diffuse + specular) * attenuation);            // Add the light values to the running total for the fragment, attenuated as required
    }

    if (TextureMap1Flag) {                                                          // If this fragment has a valid TextureMap1Unit assigned:

        vec4  texel = texture (TextureMap1Unit, texcoord.st);                       // Extract the texel at this fragment's texture coordinates

        float combinedAlpha = (texel.a * MaterialAlpha);                            // Combine the alpha values from both texel and fragment

        fragmentColor = vec4 (fragmentColor.rgb * texel.rgb, combinedAlpha);        // Provide the old GL_MODULATE equivalent fragment color
    }
    
    if (TextureMap2Flag) {                                                          // If this fragment has a valid TextureMap2Unit assigned:

        vec4  texel = texture (TextureMap2Unit, texcoord.st);                       // Extract the texel at this fragment's texture coordinates

        float combinedAlpha = (texel.a * MaterialAlpha);                            // Combine the alpha values from both texel and fragment

        fragmentColor = vec4 (fragmentColor.rgb * texel.rgb, combinedAlpha);        // Provide the old GL_MODULATE equivalent fragment color
    }

    FragColour = fragmentColor;                                                     // Provide the lit fragmentColor to next stage in OpenGL pipeline
}
