#version 150                                                            // Stipulate OpenGL 3.2

#define MAX_BONES 66                                                    // Maximum Bones (MUST match Constants.h or we lose bumpmap textures)

uniform mat4    ModelViewMatrix;                                        // Current ModelView matrix, passed in during each frame update
uniform mat4    ProjectionMatrix;                                       // Current Projection matrix, also passed in during each frame update
uniform uint    AnimatedBoneCount;                                      // Count of animated Bones; a value >0 means this vertex is to be deformed
uniform mat2x4  BoneTransforms[MAX_BONES];                              // Array of Bone transforms (Dual Quaternions), plus an empty guardrail entry

in vec3         Position;                                               // Position     = 3 floats
in vec3         Normal;                                                 // Normal       = 3 floats
in vec3         Tangent;                                                // Tangent      = 3 floats
in vec2         TexCoord;                                               // Texcoord     = 2 floats
in vec4         BoneIndices;                                            // Bone Indices = 4 floats (indices of bones deforming the vertex)
in vec4         BoneWeights;                                            // Bone Weights = 4 floats (normalized weights of deforming bones)

out vec4        position;                                               // Output sent downstream to rasterizer, fragment shader, etc.
out vec3        normal;
out vec3        tangent;
out vec2        texcoord;

mat2x4 GetBlendedBoneTransform() {                                      // From https://dev.theomader.com/dual-quaternion-skinning/
    
    mat2x4 dq1, dq2, dq3, dq4, blendedDQ;                               // Matrices of 2 columns x 4 rows, in Column-Major order
    
    uvec4 BoneIDs = uvec4(BoneIndices);                                 // Cast incoming bone indices from float to uvec4 
    
    dq1 = BoneTransforms[BoneIDs[0]];                                   // 1st bone's deformation transform as a Dual Quaternion
    dq2 = BoneTransforms[BoneIDs[1]];                                   // 2nd   "        "           "     "     "       "
    dq3 = BoneTransforms[BoneIDs[2]];                                   // 3rd   "        "           "     "     "       "
    dq4 = BoneTransforms[BoneIDs[3]];                                   // 4th   "        "           "     "     "       "
    
    if (dot (dq1[0], dq2[0]) < 0.0) dq2 *= -1.0;                        // Ensure bone transforms are in the same 'neighbourhood'
    if (dot (dq1[0], dq3[0]) < 0.0) dq3 *= -1.0;
    if (dot (dq1[0], dq4[0]) < 0.0) dq4 *= -1.0;
    
    blendedDQ = (dq1 * BoneWeights[0]) + (dq2 * BoneWeights[1]) + (dq3 * BoneWeights[2]) + (dq4 * BoneWeights[3]);

    blendedDQ[0] = normalize (blendedDQ[0]);                            // Normalize (only) the Dual Quaternion's rotation half
    
    return blendedDQ;
}

vec3 TransformPosition (vec3 position, vec4 realDQ, vec4 dualDQ) {      // See https://donw.io/post/dual-quaternion-skinning/

    return position + 2 *  cross (realDQ.xyz, cross (realDQ.xyz, position) + realDQ.w * position)
                    + 2 * (cross (realDQ.xyz, dualDQ.xyz) + (realDQ.w * dualDQ.xyz) - (dualDQ.w * realDQ.xyz));
}

vec3 TransformNormal (vec3 normal, vec4 realDQ) {

    return normal + 2.0 * cross (realDQ.xyz, cross (realDQ.xyz, normal) + realDQ.w * normal);
}


void main (void) {
    
    vec3 deformedPosition = Position;                                                   // Get some working copies of immutable inputs
    vec3 deformedNormal   = Normal;
    
    if ((AnimatedBoneCount > uint(0)) && (AnimatedBoneCount < uint(MAX_BONES))) {       // If animated bones, deform vertex and normal
        
        mat2x4 blendedDQ  = GetBlendedBoneTransform();                                  // Get blended bone transform (Dual Quaternion)

        vec4 BDQ_Real     = vec4 (blendedDQ[0][0], blendedDQ[0][1], blendedDQ[0][2], blendedDQ[0][3]);  // Extract Real part (rotation)
        vec4 BDQ_Dual     = vec4 (blendedDQ[1][0], blendedDQ[1][1], blendedDQ[1][2], blendedDQ[1][3]);  //    "    Dual   "  (translation)

        deformedPosition  = TransformPosition (deformedPosition, BDQ_Real, BDQ_Dual);   // (Model-space) vertex from bind to posed position
        deformedNormal    = TransformNormal   (deformedNormal,   BDQ_Real);             // (Model-space) normal from bind to posed normal
    }

    position    = ProjectionMatrix * ModelViewMatrix * vec4 (deformedPosition, 1.0);    // Transform position to Eye-space, then Clip-space

    normal      = vec3 (ModelViewMatrix * vec4 (normalize (deformedNormal), 0.0));      // Transform the vertex's normal vector to Eye-space...
                                                                                        // ... to give the illusion of 'Camera as light source'
    tangent     = normalize (Tangent);
    
    texcoord    = TexCoord;
    
    gl_Position = position;
}
